""" Utilities for working with SBML models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (  # noqa: F401
    SedDocument, ModelAttributeChange, Variable, Symbol,
    Simulation, OneStepSimulation, SteadyStateSimulation, UniformTimeCourseSimulation,
    Algorithm, Task,
    )
from ...utils.core import format_float, flatten_nested_list_of_strings
from .validation import validate_model
import os
import re
import types  # noqa: F401
import typing

__all__ = ['get_parameters_variables_outputs_for_simulation', 'get_package_namespace']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument,
                                                    native_ids=False, native_data_types=False,
                                                    include_local_parameters_in_task_level_simulation_parameters=False,
                                                    include_model_parameters_in_simulation_variables=False,
                                                    include_compartment_sizes_in_simulation_variables=False,
                                                    include_reaction_fluxes_in_kinetic_simulation_variables=False,
                                                    validate=True, validate_consistency=True):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:sbml``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return ``new_value`` in their native data types
        include_local_parameters_in_task_level_simulation_parameters (:obj:`bool`, optional): whether to include the local parameters
            among the returned SED model changes for SED tasks
        include_model_parameters_in_simulation_variables (:obj:`bool`, optional): whether to include the values of
            non-constant SBML parameters with assignment rules among the returned SED variables
        include_compartment_sizes_in_simulation_variables (:obj:`bool`, optional): whether to include the sizes of
            non-constant SBML compartments with assignment rules among the returned SED variables
        include_reaction_fluxes_in_kinetic_simulation_variables (:obj:`bool`, optional): whether to include reaction fluxes
            among the returned SED variables for kinetic simulations
        validate (:obj:`str`, optional): whether to validate the model
        validate_consistency (:obj:`str`, optional): whether to check the consistency of the model

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    # check model file exists
    if not os.path.isfile(model_filename):
        raise FileNotFoundError('Model file `{}` does not exist'.format(model_filename))

    # check change level
    if change_level not in [SedDocument, Task]:
        msg = 'Change level {} is not supported. Changes can only made at the SED document or task level.'.format(change_level)
        raise NotImplementedError(msg)

    # read model
    errors, _, doc = validate_model(model_filename, validate_consistency=validate and validate_consistency)
    if not doc or (validate and errors):
        raise ValueError('Model file `{}` is not a valid SBML file.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))
    model = doc.getModel()

    # determine plugin usage
    has_fbc = False
    has_qual = False
    has_other = False

    plugin_names = []
    for i_plugin in range(model.getNumPlugins()):
        plugin = model.getPlugin(i_plugin)
        plugin_name = plugin.getPackageName()
        plugin_names.append(plugin_name)
        if plugin_name == 'fbc':
            has_fbc = True
        elif plugin_name == 'qual':
            has_qual = True
        elif plugin_name not in ['layout', 'math', 'render', 'spatial', 'annot', 'req']:
            has_other = True

    if has_other or (has_fbc and has_qual):
        raise NotImplementedError('Unable to get the observable variables for a model that uses these plugins:\n  {}'.format(
            '\n  '.join(sorted(plugin_names))))

    # initialize dependent parameters and independent variables
    params = []
    vars = []

    # add time to vars
    if simulation_type in [OneStepSimulation, UniformTimeCourseSimulation]:
        if has_fbc:
            raise NotImplementedError('One step and time course simulations are not supported for FBC models')

        vars.append(Variable(
            id=None if native_ids else 'time',
            name=None if native_ids else 'Time',
            symbol=Symbol.time,
        ))

    elif simulation_type in [SteadyStateSimulation]:
        pass

    else:
        raise NotImplementedError('Simulation of type `{}` are not supported'.format(simulation_type))

    # add independent variables
    if has_fbc:
        sim = SteadyStateSimulation(
            id='simulation',
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id or 'KISAO_0000437',
            ),
        )

        plugin = model.getPlugin('fbc')

        has_flux = False
        has_flux_bounds = False

        if sim.algorithm.kisao_id in ['KISAO_0000437', 'KISAO_0000527', 'KISAO_0000528', 'KISAO_0000554']:
            # FBA, gFBA, pFBA
            has_flux = True

        elif sim.algorithm.kisao_id in ['KISAO_0000526']:
            # FVA
            has_flux_bounds = True

        else:
            raise NotImplementedError('Algorithm with KiSAO id `{}` is not supported'.format(algorithm_kisao_id))

        if change_level == SedDocument:
            namespaces = {
                'sbml': model.getURI(),
            }
            for parameter in model.getListOfParameters():
                param_id = parameter.getId()

                if not model.getInitialAssignmentBySymbol(param_id) and not model.getAssignmentRuleByVariable(param_id):
                    params.append(ModelAttributeChange(
                        id=param_id if native_ids else 'value_parameter_' + param_id,
                        name=(
                            parameter.getName() or None
                            if native_ids else
                            'Value of parameter "{}"'.format(parameter.getName() or param_id)
                        ),
                        target="/sbml:sbml/sbml:model/sbml:listOfParameters/sbml:parameter[@id='{}']/@value".format(param_id),
                        target_namespaces=namespaces,
                        new_value=parameter.getValue() if native_data_types else format_float(parameter.getValue()),
                    ))

        elif change_level == Task:
            namespaces = {
                'sbml': model.getURI(),
                'fbc': plugin.getURI(),
            }
            for reaction in model.getListOfReactions():
                reaction_fbc = reaction.getPlugin('fbc')
                lower_flux_bound = reaction_fbc.getLowerFluxBound()
                upper_flux_bound = reaction_fbc.getUpperFluxBound()
                lower_flux_bound_param = model.getParameter(lower_flux_bound)
                upper_flux_bound_param = model.getParameter(upper_flux_bound)
                lower_flux_bound = lower_flux_bound_param.getValue()
                upper_flux_bound = upper_flux_bound_param.getValue()
                if not native_data_types:
                    lower_flux_bound = str(lower_flux_bound)
                    upper_flux_bound = str(upper_flux_bound)

                rxn_id = reaction.getId()
                if (
                    not model.getInitialAssignmentBySymbol(lower_flux_bound_param.getId())
                    and not model.getAssignmentRuleByVariable(lower_flux_bound_param.getId())
                ):
                    params.append(ModelAttributeChange(
                        id=rxn_id if native_ids else 'lower_bound_reaction_' + rxn_id,
                        name=(
                            reaction.getName() or None
                            if native_ids else
                            'Lower bound of reaction "{}"'.format(reaction.getName() or rxn_id)
                        ),
                        target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@fbc:lowerFluxBound".format(rxn_id),
                        target_namespaces=namespaces,
                        new_value=lower_flux_bound,
                    ))
                if (
                    not model.getInitialAssignmentBySymbol(upper_flux_bound_param.getId())
                    and not model.getAssignmentRuleByVariable(upper_flux_bound_param.getId())
                ):
                    params.append(ModelAttributeChange(
                        id=rxn_id if native_ids else 'upper_bound_reaction_' + rxn_id,
                        name=(
                            reaction.getName() or None
                            if native_ids else
                            'Upper bound of reaction "{}"'.format(reaction.getName() or rxn_id)
                        ),
                        target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@fbc:upperFluxBound".format(rxn_id),
                        target_namespaces=namespaces,
                        new_value=upper_flux_bound,
                    ))

        if has_flux:
            namespaces = {
                'sbml': model.getURI(),
                'fbc': plugin.getURI(),
            }
            obj = plugin.getActiveObjective()
            obj_id = obj.getId()
            var = Variable(
                id=obj_id if native_ids else 'value_objective_' + obj_id,
                name=obj.getName() or None if native_ids else 'Value of objective "{}"'.format(obj.getName() or obj_id),
                target="/sbml:sbml/sbml:model/fbc:listOfObjectives/fbc:objective[@fbc:id='{}']/@value".format(obj_id),
                target_namespaces=namespaces,
            )
            vars.append(var)

            namespaces = {
                'sbml': model.getURI(),
            }
            for reaction in model.getListOfReactions():
                rxn_id = reaction.getId()
                var = Variable(
                    id=rxn_id if native_ids else 'flux_reaction_' + rxn_id,
                    name=reaction.getName() or None if native_ids else 'Flux of reaction "{}"'.format(reaction.getName() or rxn_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@flux".format(rxn_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

        if has_flux_bounds:
            namespaces = {
                'sbml': model.getURI(),
            }

            for reaction in model.getListOfReactions():
                rxn_id = reaction.getId()

                var = Variable(
                    id=rxn_id if native_ids else 'min_flux_reaction_' + rxn_id,
                    name=reaction.getName() or None if native_ids else 'Minimum flux of reaction "{}"'.format(reaction.getName() or rxn_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@minFlux".format(rxn_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

                var = Variable(
                    id=rxn_id if native_ids else 'max_flux_reaction_' + rxn_id,
                    name=reaction.getName() or None if native_ids else 'Maximum flux of reaction "{}"'.format(reaction.getName() or rxn_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@maxFlux".format(rxn_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

    elif has_qual:
        if simulation_type == OneStepSimulation:
            sim = OneStepSimulation(
                step=1.,
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000449',
                ),
            )
        elif simulation_type == SteadyStateSimulation:
            sim = SteadyStateSimulation(
                id='simulation',
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000659',
                ),
            )
        else:
            sim = UniformTimeCourseSimulation(
                id='simulation',
                initial_time=0.,
                output_start_time=0.,
                output_end_time=10.,
                number_of_steps=10,
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000449',
                ),
            )

        plugin = model.getPlugin('qual')
        namespaces = {
            'sbml': model.getURI(),
            'qual': plugin.getURI(),
        }
        for species in plugin.getListOfQualitativeSpecies():
            species_id = species.getId()

            if species.isSetInitialLevel():
                if not model.getInitialAssignmentBySymbol(species_id) and not model.getAssignmentRuleByVariable(species_id):
                    params.append(ModelAttributeChange(
                        id=species_id if native_ids else 'init_level_species_{}'.format(species_id),
                        name=(
                            species.getName() or None
                            if native_ids else
                            'Initial level of species "{}"'.format(species.getName() or species_id)
                        ),
                        target=(
                            "/sbml:sbml/sbml:model/qual:listOfQualitativeSpecies"
                            "/qual:qualitativeSpecies[@qual:id='{}']/@qual:initialLevel"
                        ).format(species_id),
                        target_namespaces=namespaces,
                        new_value=species.getInitialLevel() if native_data_types else str(species.getInitialLevel()),
                    ))

            if not species.isSetConstant() or not species.getConstant():
                var = Variable(
                    id=species_id if native_ids else 'level_species_' + species_id,
                    name=species.getName() or None if native_ids else 'Level of species "{}"'.format(species.getName() or species_id),
                    target="/sbml:sbml/sbml:model/qual:listOfQualitativeSpecies/qual:qualitativeSpecies[@qual:id='{}']".format(species_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

        namespaces = {
            'sbml': model.getURI(),
        }
        for comp in model.getListOfCompartments():
            comp_id = comp.getId()

            if comp.isSetSize() and (not model.getInitialAssignmentBySymbol(comp_id) and not model.getAssignmentRuleByVariable(comp_id)):
                params.append(ModelAttributeChange(
                    id=comp.getId() if native_ids else 'init_size_compartment_{}'.format(comp.getId()),
                    name=(
                        comp.getName() or None
                        if native_ids else
                        'Initial size of compartment "{}"'.format(comp.getName() or comp.getId())
                    ),
                    target="/sbml:sbml/sbml:model/sbml:listOfCompartments/sbml:compartment[@id='{}']/@size".format(comp_id),
                    target_namespaces=namespaces,
                    new_value=comp.getSize() if native_data_types else format_float(comp.getSize()),
                ))

            if (
                include_compartment_sizes_in_simulation_variables
                and comp.isSetConstant()
                and not comp.getConstant()
                and model.getAssignmentRuleByVariable(comp_id)
            ):
                var = Variable(
                    id=comp_id if native_ids else 'size_compartment_' + comp_id,
                    name=comp.getName() or None if native_ids else 'Size of compartment "{}"'.format(comp.getName() or comp_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfCompartments/sbml:compartment[@id='{}']/@size".format(comp_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

    else:
        if simulation_type == OneStepSimulation:
            sim = OneStepSimulation(
                id='simulation',
                step=1.,
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000019',
                ),
            )
        elif simulation_type == SteadyStateSimulation:
            sim = SteadyStateSimulation(
                id='simulation',
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000408',
                ),
            )
        else:
            sim = UniformTimeCourseSimulation(
                id='simulation',
                initial_time=0.,
                output_start_time=0.,
                output_end_time=1.,
                number_of_steps=10,
                algorithm=Algorithm(
                    kisao_id=algorithm_kisao_id or 'KISAO_0000019',
                ),
            )

        namespaces = {
            'sbml': model.getURI(),
        }
        for species in model.getListOfSpecies():
            species_id = species.getId()

            if not model.getInitialAssignmentBySymbol(species_id) and not model.getAssignmentRuleByVariable(species_id):
                if species.isSetInitialAmount():
                    params.append(ModelAttributeChange(
                        id=species_id if native_ids else 'init_amount_species_{}'.format(species_id),
                        name=(
                            species.getName() or None
                            if native_ids else
                            'Initial amount of species "{}"'.format(species.getName() or species_id)
                        ),
                        target="/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species[@id='{}']/@initialAmount".format(species_id),
                        target_namespaces=namespaces,
                        new_value=species.getInitialAmount() if native_data_types else format_float(species.getInitialAmount()),
                    ))
                elif species.isSetInitialConcentration():
                    params.append(ModelAttributeChange(
                        id=species_id if native_ids else 'init_conc_species_{}'.format(species_id),
                        name=(
                            species.getName() or None
                            if native_ids else
                            'Initial concentration of species "{}"'.format(species.getName() or species_id)
                        ),
                        target="/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species[@id='{}']/@initialConcentration".format(species_id),
                        target_namespaces=namespaces,
                        new_value=(
                            species.getInitialConcentration()
                            if native_data_types else
                            format_float(species.getInitialConcentration())
                        ),
                    ))

            if not species.isSetConstant() or not species.getConstant():
                var = Variable(
                    id=species_id if native_ids else 'dynamics_species_' + species_id,
                    name=species.getName() or None if native_ids else 'Dynamics of species "{}"'.format(species.getName() or species_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species[@id='{}']".format(species_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

        for comp in model.getListOfCompartments():
            comp_id = comp.getId()

            if (
                comp.isSetSize()
                and not model.getInitialAssignmentBySymbol(comp_id)
                and not model.getAssignmentRuleByVariable(comp_id)
            ):
                params.append(ModelAttributeChange(
                    id=(
                        comp.getId()
                        if native_ids else
                        'init_size_compartment_{}'.format(comp.getId())
                    ),
                    name=(
                        comp.getName() or None
                        if native_ids else
                        'Initial size of compartment "{}"'.format(comp.getName() or comp.getId())
                    ),
                    target="/sbml:sbml/sbml:model/sbml:listOfCompartments/sbml:compartment[@id='{}']/@size".format(comp_id),
                    target_namespaces=namespaces,
                    new_value=comp.getSize() if native_data_types else format_float(comp.getSize()),
                ))

            if (
                include_compartment_sizes_in_simulation_variables
                and comp.isSetConstant()
                and not comp.getConstant()
                and model.getAssignmentRuleByVariable(comp_id)
            ):
                var = Variable(
                    id=comp_id if native_ids else 'size_compartment_' + comp_id,
                    name=comp.getName() or None if native_ids else 'Size of compartment "{}"'.format(comp.getName() or comp_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfCompartments/sbml:compartment[@id='{}']/@size".format(comp_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

        for parameter in model.getListOfParameters():
            param_id = parameter.getId()

            if not model.getInitialAssignmentBySymbol(param_id) and not model.getAssignmentRuleByVariable(param_id):
                params.append(ModelAttributeChange(
                    id=param_id if native_ids else 'value_parameter_' + param_id,
                    name=parameter.getName() or None if native_ids else 'Value of parameter "{}"'.format(parameter.getName() or param_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfParameters/sbml:parameter[@id='{}']/@value".format(param_id),
                    target_namespaces=namespaces,
                    new_value=parameter.getValue() if native_data_types else format_float(parameter.getValue()),
                ))

            if (
                include_model_parameters_in_simulation_variables
                and parameter.isSetConstant()
                and not parameter.getConstant()
                and model.getAssignmentRuleByVariable(param_id)
            ):
                var = Variable(
                    id=param_id if native_ids else 'value_parameter_' + param_id,
                    name=parameter.getName() or None if native_ids else 'Value of parameter "{}"'.format(parameter.getName() or param_id),
                    target="/sbml:sbml/sbml:model/sbml:listOfParameters/sbml:parameter[@id='{}']/@value".format(param_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

        for reaction in model.getListOfReactions():
            reaction_id = reaction.getId()
            kinetic_law = reaction.getKineticLaw()
            if kinetic_law and (change_level == SedDocument or include_local_parameters_in_task_level_simulation_parameters):
                for parameter in kinetic_law.getListOfParameters():
                    param_id = parameter.getId()

                    if model.getLevel() >= 3:
                        target = (
                            "/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/sbml:kineticLaw"
                            "/sbml:listOfLocalParameters/sbml:localParameter[@id='{}']/@value"
                        ).format(reaction_id, param_id)
                    else:
                        target = (
                            "/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/sbml:kineticLaw"
                            "/sbml:listOfParameters/sbml:parameter[@id='{}']/@value"
                        ).format(reaction_id, param_id)

                    if not model.getInitialAssignmentBySymbol(param_id) and not model.getAssignmentRuleByVariable(param_id):
                        params.append(ModelAttributeChange(
                            id=param_id if native_ids else 'value_parameter_' + param_id,
                            name=parameter.getName() or None if native_ids else 'Value of parameter "{}" of reaction "{}"'.format(
                                parameter.getName() or param_id, reaction.getName() or reaction_id),
                            target=target,
                            target_namespaces=namespaces,
                            new_value=parameter.getValue() if native_data_types else format_float(parameter.getValue()),
                        ))

                    if (
                        include_model_parameters_in_simulation_variables
                        and parameter.isSetConstant()
                        and not parameter.getConstant()
                        and model.getAssignmentRuleByVariable(param_id)
                    ):
                        var = Variable(
                            id=param_id if native_ids else 'value_parameter_' + param_id,
                            name=parameter.getName() or None if native_ids else 'Value of parameter "{}" of reaction "{}"'.format(
                                parameter.getName() or param_id, reaction.getName() or reaction_id),
                            target=target,
                            target_namespaces=namespaces,
                        )
                        vars.append(var)

            if include_reaction_fluxes_in_kinetic_simulation_variables:
                var = Variable(
                    id=(
                        reaction_id
                        if native_ids else
                        'flux_reaction_' + reaction_id
                    ),
                    name=(
                        reaction.getName() or None
                        if native_ids else
                        'Flux of reaction "{}"'.format(reaction.getName() or reaction_id)
                    ),
                    target="/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']".format(reaction_id),
                    target_namespaces=namespaces,
                )
                vars.append(var)

    return (params, [sim], vars, [])


def get_package_namespace(package: str, namespaces: typing.Dict[typing.Union[str, None], str]) -> typing.Tuple[str, str]:
    """ Get the prefix and URI for the namespace of an SBML package

    Args:
        package (:obj:`str`): SBML package id (e.g., ``fbc``, ``qual``)
        namespaces (:obj:`dict`): dictionary that maps namespace prefixes to URIs

    Returns:
        :obj:`tuple`:

            * :obj:`str`: prefix
            * :obj:`str`: URI
    """
    sbml_fbc_prefixes = set()
    sbml_fbc_uris = set()
    for prefix, uri in namespaces.items():
        if uri and re.match(r'^http://www.sbml.org/sbml/level\d+/version\d+/{}/version\d+$'.format(package), uri):
            sbml_fbc_prefixes.add(prefix)
            sbml_fbc_uris.add(uri)
    if len(sbml_fbc_uris) != 1:
        raise ValueError('Namespaces must include 1 SBML {} namespace.'.format(package))
    return list(sbml_fbc_prefixes)[0], list(sbml_fbc_uris)[0]



""" Utilities for validating Smoldyn models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-13
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

# TODO: remove once Smoldyn updated to BioSimulators-utils >= 0.1.96
import biosimulators_utils.viz
import biosimulators_utils.viz.data_model
import biosimulators_utils.viz.io
import biosimulators_utils.viz.warnings
import sys
sys.modules['biosimulators_utils.plot'] = sys.modules['biosimulators_utils.viz']
sys.modules['biosimulators_utils.plot.data_model'] = sys.modules['biosimulators_utils.viz.data_model']
sys.modules['biosimulators_utils.plot.io'] = sys.modules['biosimulators_utils.viz.io']
sys.modules['biosimulators_utils.plot.warnings'] = sys.modules['biosimulators_utils.viz.warnings']
biosimulators_utils.viz.data_model.PlotFormat = biosimulators_utils.viz.data_model.VizFormat

from ...log.data_model import StandardOutputErrorCapturerLevel  # noqa: E402
from ...log.utils import StandardOutputErrorCapturer  # noqa: E402
from smoldyn.biosimulators.combine import (  # noqa: E402
    read_smoldyn_simulation_configuration,
    disable_smoldyn_graphics_in_simulation_configuration,
    write_smoldyn_simulation_configuration,
    init_smoldyn_simulation_from_configuration_file,
)
import os  # noqa: E402
import tempfile  # noqa: E402


def validate_model(filename, name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`tuple`:

                * :obj:`smoldyn.Simulation`: model configuration
                * :obj:`list` of :obj:`str`: model configuration
    """
    errors = []
    warnings = []
    model = None
    config = None

    if filename:
        if os.path.isfile(filename):
            config = read_smoldyn_simulation_configuration(filename)
            disable_smoldyn_graphics_in_simulation_configuration(config)
            fid, config_filename = tempfile.mkstemp(suffix='.txt')
            os.close(fid)
            write_smoldyn_simulation_configuration(config, config_filename)
            with StandardOutputErrorCapturer(level=StandardOutputErrorCapturerLevel.c, relay=False) as captured:
                try:
                    model = init_smoldyn_simulation_from_configuration_file(config_filename)
                    valid = True
                except ValueError:
                    valid = False
            if not valid:
                errors.append(['`{}` is not a valid Smoldyn configuration file.'.format(filename), [[captured.get_text()]]])
            os.remove(config_filename)

        else:
            errors.append(['`{}` is not a file.'.format(filename or '')])

    else:
        errors.append(['`filename` must be a path to a file, not `{}`.'.format(filename or '')])

    return (errors, warnings, (model, config))



""" Utilities for working with COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..combine.data_model import CombineArchive  # noqa: F401
from ..combine.utils import get_sedml_contents
from ..sedml.data_model import SedDocument, Task, Output, Report, Plot2D, Plot3D, DataSet, Curve, Surface
from ..sedml.io import SedmlSimulationReader
from ..warnings import warn
from .data_model import (Status, CombineArchiveLog, SedDocumentLog,  # noqa: F401
                         TaskLog, OutputLog, ReportLog, Plot2DLog, Plot3DLog,
                         StandardOutputErrorCapturerLevel)
from .warnings import StandardOutputNotLoggedWarning
try:
    import capturer
except ModuleNotFoundError:
    capturer = None
import contextlib
import io  # noqa: F401
import os
import sys

__all__ = [
    'init_combine_archive_log',
    'init_sed_document_log',
    'init_task_log',
    'init_output_log',
    'init_report_log',
    'init_plot2d_log',
    'init_plot3d_log',
    'StandardOutputErrorCapturer',
    'get_summary_combine_archive_log',
]


def init_combine_archive_log(archive, archive_dir,
                             supported_features=(SedDocument, Task, Report, Plot2D, Plot3D, DataSet, Curve, Surface),
                             logged_features=(SedDocument, Task, Report, Plot2D, Plot3D, DataSet, Curve, Surface)):
    """ Initialize a log of a COMBINE/OMEX archive

    Args:
        archive (:obj:`CombineArchive`): COMBINE/OMEX archive
        archive_dir (:obj:`str`): path where the content of the archive is located
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: COMBINE/OMEX archives and SED documents, tasks, reports, plots,
            data sets, curves, and surfaces.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: COMBINE/OMEX archives and SED documents, tasks, reports, plots,
            data sets, curves, and surfaces.

    Returns:
        :obj:`CombineArchiveLog`: initialized log of a COMBINE/OMEX archive
    """
    contents = get_sedml_contents(archive)

    log = CombineArchiveLog(status=Status.QUEUED)

    if SedDocument in logged_features:
        log.sed_documents = {}
        for content in contents:
            content_filename = os.path.join(archive_dir, content.location)
            doc = SedmlSimulationReader().run(content_filename, validate_semantics=False, validate_models_with_languages=False)

            doc_log = init_sed_document_log(doc, supported_features=supported_features, logged_features=logged_features)
            doc_log.location = os.path.relpath(content.location, '.')
            doc_log.status = Status.QUEUED if isinstance(doc, supported_features) else Status.SKIPPED

            doc_log.parent = log
            doc_id = os.path.relpath(content_filename, archive_dir)
            log.sed_documents[doc_id] = doc_log

    else:
        log.sed_documents = None

    return log


def init_sed_document_log(doc,
                          supported_features=(Task, Report, Plot2D, Plot3D, DataSet, Curve, Surface),
                          logged_features=(Task, Report, Plot2D, Plot3D, DataSet, Curve, Surface)):
    """ Initialize a log of a SED document

    Args:
        doc (:obj:`SedDocument`): SED document
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: tasks, reports, plots, data sets, curves, and surfaces.
        logged_features (:obj:`list` of :obj:`type`, optional): list of SED elements which
            will be logged. Default: tasks, reports, plots, data sets, curves, and surfaces.

    Returns:
        :obj:`SedDocumentLog`: initialized log of a SED document
    """
    log = SedDocumentLog()

    if Task in logged_features:
        log.tasks = {}
        for task in doc.tasks:
            task_log = init_task_log(task, supported_features=supported_features, logged_features=logged_features)
            task_log.status = Status.QUEUED if isinstance(task, supported_features) else Status.SKIPPED
            task_log.parent = log
            log.tasks[task.id] = task_log
    else:
        log.tasks = None

    if set([Output, Report, Plot2D, Plot3D]).intersection(logged_features):
        log.outputs = {}

        for output in doc.outputs:
            if isinstance(output, logged_features):
                output_log = init_output_log(output, supported_features=supported_features, logged_features=logged_features)
                output_log.status = Status.QUEUED if isinstance(output, supported_features) else Status.SKIPPED
                output_log.parent = log
                log.outputs[output.id] = output_log

    else:
        log.outputs = None

    return log


def init_task_log(task,
                  supported_features=(),
                  logged_features=()):
    """ Initialize a log of a task

    Args:
        output (:obj:`Task`): a SED task
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: empty list.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: empty list.

    Returns:
        :obj:`OutputLog`: initialized log of a SED document
    """
    return TaskLog(id=task.id)


def init_output_log(output,
                    supported_features=(DataSet, Curve, Surface),
                    logged_features=(DataSet, Curve, Surface)):
    """ Initialize a log of an output

    Args:
        output (:obj:`Output`): a SED output
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: data sets, curves, and surfaces.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: data sets, curves, and surfaces.

    Returns:
        :obj:`OutputLog`: initialized log of a SED document
    """

    if isinstance(output, Report):
        log = init_report_log(output, supported_features=supported_features, logged_features=logged_features)

    elif isinstance(output, Plot2D):
        log = init_plot2d_log(output, supported_features=supported_features, logged_features=logged_features)

    elif isinstance(output, Plot3D):
        log = init_plot3d_log(output, supported_features=supported_features, logged_features=logged_features)

    else:
        raise NotImplementedError('`{}` outputs are not supported.'.format(
            output.__class__.__name__))  # pragma: no cover # unreachable because all cases are enumerated above

    return log


def init_report_log(report,
                    supported_features=(DataSet, Curve, Surface),
                    logged_features=(DataSet, Curve, Surface)):
    """ Initialize a log of a report

    Args:
        report (:obj:`Report`): a SED report
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: data sets.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: data sets.

    Returns:
        :obj:`ReportLog`: initialized log of a report
    """

    log = ReportLog(id=report.id)

    if DataSet in logged_features:
        log.data_sets = {}
        for data_set in report.data_sets:
            log.data_sets[data_set.id] = (
                Status.QUEUED
                if isinstance(data_set, supported_features)
                else Status.SKIPPED)
    else:
        log.data_sets = None

    return log


def init_plot2d_log(plot,
                    supported_features=(Curve),
                    logged_features=(Curve)):
    """ Initialize a log of a 2D plot

    Args:
        plot (:obj:`Plot2D`): a SED 2D plot
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: curves.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: curves.

    Returns:
        :obj:`Plot2DLog`: initialized log of a 2D plot
    """
    log = Plot2DLog(id=plot.id)

    if Curve in logged_features:
        log.curves = {}
        for curve in plot.curves:
            log.curves[curve.id] = (
                Status.QUEUED
                if isinstance(curve, supported_features)
                else Status.SKIPPED)
    else:
        log.curves = None

    return log


def init_plot3d_log(plot,
                    supported_features=(Surface),
                    logged_features=(Surface)):
    """ Initialize a log of a 3D plot

    Args:
        plot (:obj:`Plot3D`): a SED 3D plot
        supported_features (:obj:`list` of :obj:`type`, optional): list of supported elements.
            Default: surfaces.
        logged_features (:obj:`list` of :obj:`type`, optional): list of elements which
            will be logged. Default: surfaces.

    Returns:
        :obj:`Plot3DLog`: initialized log of a 3D plot
    """
    log = Plot3DLog(id=plot.id)

    if Surface in logged_features:
        log.surfaces = {}
        for surface in plot.surfaces:
            log.surfaces[surface.id] = (
                Status.QUEUED
                if isinstance(surface, supported_features)
                else Status.SKIPPED)
    else:
        log.surfaces = None

    return log


class StandardOutputErrorCapturer(contextlib.AbstractContextManager):
    """ Context manager for capturing standard output/error. When :obj:`capturer` is available (i.e.,
    Linux, MacOS, Unix), :obj:`capturer` is used to capture standard output/error. When :obj:`capturer` is not
    available (i.e. Windows), this context manager issues a warn and collects no output. The purpose of this
    context manager is to encapsulate the handling of whether :obj:`capturer` is or isn't available so
    that the other modules can work seamless in Linux, as well as Windows (except without the ability to log
    standard output/error).

    Attributes:
        level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which stdout/stderr should be captured
        relay (:obj:`bool`): if :obj:`True`, collect the standard output/error streams and continue to pass
                them along. if :obj:`False`, collect the stream, squash them, and do not pass them along.
        disabled (:obj:`bool`): whether to capture standard output and error
        _captured (:obj:`capturer.CaptureOutput`): logged C output
        _log (:obj:`str`): logged Python output
        _stdout (:obj:`io.IOBase`): overridden stdout
        _stderr (:obj:`io.IOBase`): overridden stderr
    """

    def __init__(self, level=StandardOutputErrorCapturerLevel.c, relay=False, termination_delay=0.01, disabled=False):
        """
        Args:
            level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which stdout/stderr should be captured
            relay (:obj:`bool`): if :obj:`True`, collect the standard output/error streams and continue to pass
                them along. if :obj:`False`, collect the stream, squash them, and do not pass them along.
            termination_delay (:obj:`float`, optional): The number of seconds to wait before terminating
                the output relay process.
            disabled (:obj:`bool`, optional): whether to capture standard output and error
        """
        self.level = level
        self.relay = relay
        self.disabled = disabled
        if not self.disabled:
            if self.level >= StandardOutputErrorCapturerLevel.c and capturer:
                self._captured = capturer.CaptureOutput(merged=True, relay=relay, termination_delay=termination_delay)
            else:
                self._log = ''
        else:
            msg = (
                'Standard output and error could not be logged because capturer is not installed. '
                'To install capturer, install BioSimulators utils with the `logging` option '
                '(`pip install biosimulators-utils[logging]`).'
            )
            warn(msg, StandardOutputNotLoggedWarning)

    def __enter__(self):
        """ Enter a context """
        if not self.disabled:
            if self.level >= StandardOutputErrorCapturerLevel.c and capturer:
                self._captured.start_capture()
            else:
                self._stdout = sys.stdout
                self._stderr = sys.stderr
                sys.stdout = self
                sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """ Exit a context """
        if not self.disabled:
            if self.level >= StandardOutputErrorCapturerLevel.c and capturer:
                self._captured.finish_capture()
            else:
                sys.stdout = self._stdout
                sys.stderr = self._stderr

    def write(self, message):
        if self.relay:
            self._stdout.write(message)
        self._log += message

    def flush(self):
        if self.relay:
            self._stdout.flush()

    def get_text(self):
        """ Get the captured standard output/error

        Returns:
            :obj:`str`: captured standard output/error
        """
        if self.disabled:
            return None
        else:
            if self.level >= StandardOutputErrorCapturerLevel.c and capturer:
                bytes = self._captured.get_bytes()
                return bytes.decode(errors='ignore')
            else:
                return self._log


def get_summary_combine_archive_log(log):
    """ Get a summary of the log of a COMBINE/OMEX archive

    Args:
        log (:obj:`CombineArchiveLog`): log of a COMBINE/OMEX archive

    Returns:
        :obj:`str`: summary of the log
    """
    tasks_logged = False
    outputs_logged = False

    n_archives = 0
    n_tasks = 0
    n_outputs = 0

    sed_doc_status_count = {
        Status.SUCCEEDED: 0,
        Status.SKIPPED: 0,
        Status.FAILED: 0,
        None: 0,
    }
    task_status_count = {
        Status.SUCCEEDED: 0,
        Status.SKIPPED: 0,
        Status.FAILED: 0,
        None: 0,
    }
    output_status_count = {
        Status.SUCCEEDED: 0,
        Status.SKIPPED: 0,
        Status.FAILED: 0,
        None: 0,
    }
    for doc_log in log.sed_documents.values():
        n_archives += 1
        sed_doc_status_count[doc_log.status] += 1
        if doc_log.tasks is not None:
            tasks_logged = True
            for task_log in doc_log.tasks.values():
                n_tasks += 1
                task_status_count[task_log.status if task_log else None] += 1
        if doc_log.outputs is not None:
            outputs_logged = True
            for output_log in doc_log.outputs.values():
                n_outputs += 1
                output_status_count[output_log.status if output_log else None] += 1

    msg = ''
    msg += 'Executed {} SED documents:\n'.format(n_archives)
    msg += '  SED documents ({}):\n'.format(n_archives)
    msg += '    Succeeded: {}\n'.format(sed_doc_status_count[Status.SUCCEEDED])
    msg += '    Skipped: {}\n'.format(sed_doc_status_count[Status.SKIPPED])
    msg += '    Failed: {}\n'.format(sed_doc_status_count[Status.FAILED])

    if tasks_logged:
        msg += '  Tasks ({}):\n'.format(n_tasks)
        msg += '    Succeeded: {}\n'.format(task_status_count[Status.SUCCEEDED])
        msg += '    Skipped: {}\n'.format(task_status_count[Status.SKIPPED])
        msg += '    Failed: {}\n'.format(task_status_count[Status.FAILED])
        if task_status_count[None]:
            msg += '    Unknown: {}\n'.format(task_status_count[None])

    if outputs_logged:
        msg += '  Outputs ({}):\n'.format(n_outputs)
        msg += '    Succeeded: {}\n'.format(output_status_count[Status.SUCCEEDED])
        msg += '    Skipped: {}\n'.format(output_status_count[Status.SKIPPED])
        msg += '    Failed: {}\n'.format(output_status_count[Status.FAILED])
        if output_status_count[None]:
            msg += '    Unknown: {}\n'.format(output_status_count[None])

    return msg



""" Utility for converting activity SBGN process description (PD) maps to Vega format

:Author: Andrew Freiburger <afreiburger@uvic.ca>
:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-08-11
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from math import inf
from xml.etree import ElementTree
import copy
import json
import os
import pkg_resources


__all__ = ['sbgn_pd_map_to_vega']


def sbgn_pd_map_to_vega(glyph_values_data_set, sbgn_filename, vega_output_filename,
                        max_width=700, max_height=700, indent=2):
    """ Convert a process description (PD) map from SBGN to Vega format.

    Args:
        glyph_values_data_set (:obj:`dict`): Vega data set with the predicted dynamics of one or more glyphs. The dictionary should
            have one of the following schemas

            * Contain a key ``values`` whose value is a list of the predicted values of the glyphs. Each element of the list
              should be a dictionary with two keys

                * ``label``: value should be the SBGN ``label`` of the corresponding glyph
                * ``values``: array of the predicted values of the glyph

            * Contain a key ``url`` whose value is a URL where the predicted values of the glyphs can be retrieved in JSON
              format. The data should be in the same format outlined above.
            * Contain a key ``sedmlUri`` whose value is an array which contains two elements that indicate

                * The location of the SED-ML file (within its parent COMBINE/OMEX archive) which will produce
                  the predicted values of the glyphs
                * The id of the SED-ML report which will contain the predicted values of the glyphs for the visualization

        sbgn_filename (:obj:`str`): the name of the SBGN file that will be parsed
        vega_output_filename (:obj:`str`): path to the ginsim simulation data
        max_width (:obj:`int`): maximum diagram width
        max_height (:obj:`int`): maximum diagram height
        indent (:obj:`int` or :obj:`None`, optional): indentation for exported Vega file
    """

    # read the SBGN file
    root = ElementTree.parse(sbgn_filename).getroot()

    # get the default namespace used in the SBGN file
    sbgn_ns = root.tag[1:].partition('}')[0]

    # get the process description map
    map = root.find('{{{}}}map'.format(sbgn_ns))

    # ============================================== PARSE THE GLYPHS (NODES) ===============================================
    glyphs = []
    for glyph_xml in map.findall('{{{}}}glyph'.format(sbgn_ns)):
        # determine the glyph process
        glyph_class = glyph_xml.attrib['class']

        # determine the glyph id and label
        glyph_id = glyph_xml.attrib['id']

        label_xml = glyph_xml.find('{{{}}}label'.format(sbgn_ns))
        if label_xml is not None:
            glyph_label = label_xml.attrib['text']
        else:
            glyph_label = None

        # determine the glyph coordinates and size
        glyph_bounding_box = glyph_xml.find('{{{}}}bbox'.format(sbgn_ns))
        x_coordinate = float(glyph_bounding_box.attrib['x'])
        y_coordinate = float(glyph_bounding_box.attrib['y'])
        glyph_width = float(glyph_bounding_box.attrib['w'])
        glyph_height = float(glyph_bounding_box.attrib['h'])
        glyph_orientation = glyph_xml.attrib.get('orientation', None)

        # organize the aforementioned data
        glyph = {
            'id': glyph_id,
            'label': glyph_label,
            'class': glyph_class,
            'orientation': glyph_orientation,
            'x': x_coordinate,
            'y': y_coordinate,
            'width': glyph_width,
            'height': glyph_height,
        }
        glyphs.append(glyph)

        # parse the sub-glyph port objects
        """
        port_values = child.findall('{{{}}}port'.format(sbgn_ns))
        if port_values:
            for port in port_values:
                port_y = port.attrib['y']
                port_x = port.attrib['x']
                port_id = port.attrib['id']

                # organize the aforementioned data
                glyphs.append({
                    'id': port_id,
                    'process': glyph_class,
                    'x': port_x,
                    'y': port_y,
                    'width': 10,
                    'height': 10,
                })
        """

    # determine the maximum boundaries of the figure
    min_x = inf
    min_y = inf
    max_x = -inf
    max_y = -inf
    for glyph in glyphs:
        min_x = min(min_x, glyph['x'] - glyph['width'] / 2)
        max_x = max(max_x, glyph['x'] + glyph['width'] / 2)
        min_y = min(min_y, glyph['y'] - glyph['height'] / 2)
        max_y = max(max_y, glyph['y'] + glyph['height'] / 2)

    min_x += 15
    max_x += 40
    min_y += 15
    max_y += 20

    sbgn_width = max_x - min_x
    sbgn_height = max_y - min_y

    if sbgn_width / max_width > sbgn_height / max_height:
        width = max_width
        coordinate_scale = width / sbgn_width
        height = sbgn_height * coordinate_scale
    else:
        height = max_height
        coordinate_scale = height / sbgn_height
        width = sbgn_width * coordinate_scale

    # redefine the glyph values
    for glyph in glyphs:
        glyph["x"] = (glyph['x'] - min_x) * coordinate_scale
        glyph["y"] = (glyph['y'] - min_y) * coordinate_scale
        glyph["height"] = glyph["height"] * coordinate_scale
        glyph["width"] = glyph["width"] * coordinate_scale

    # ============================================== PARSE THE ARCS (EDGES) ===============================================

    # parse the arcs
    arcs = []
    for child in map.findall('{{{}}}arc'.format(sbgn_ns)):
        # determine the arc process and general information
        arc_id = child.attrib['id']
        arc_class = child.attrib['class']
        source_glyph = child.attrib['source']
        target_glyph = child.attrib['target']

        # determine the coordinates of the arc
        start_point = child.find('{{{}}}start'.format(sbgn_ns))
        x1 = (float(start_point.attrib['x']) - min_x) * coordinate_scale
        y1 = (float(start_point.attrib['y']) - min_y) * coordinate_scale

        end_point = child.find('{{{}}}end'.format(sbgn_ns))
        x2 = (float(end_point.attrib['x']) - min_x) * coordinate_scale
        y2 = (float(end_point.attrib['y']) - min_y) * coordinate_scale

        # organizing the arc information
        arcs.append({
            "id": arc_id,
            'class': arc_class,
            'source': source_glyph,
            'target': target_glyph,
            'x1': x1,
            'y1': y1,
            'x2': x2,
            'y2': y2,
        })

    # ============================================== EXPORT TO VEGA ===============================================
    # import the Vega template
    vega_template_filename = pkg_resources.resource_filename(
        'biosimulators_utils',
        os.path.join('viz', 'vega', 'sbgn', 'pd.template.json'))
    with open(vega_template_filename, 'r') as file:
        vega = json.load(file)
    vega['width'] = width
    vega['height'] = height

    # assign values to the Vega data sets
    signal_name_to_signal = {signal['name']: signal for signal in vega['signals']}
    signal_name_to_signal['sbgnWidth']['value'] = width
    signal_name_to_signal['sbgnHeight']['value'] = height

    signal = signal_name_to_signal['OutputStartTime']
    signal.pop('value', None)
    signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'outputStartTime']

    signal = signal_name_to_signal['OutputEndTime']
    signal.pop('value', None)
    signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'outputEndTime']

    signal = signal_name_to_signal['NumberOfSteps']
    signal.pop('value', None)
    signal['sedmlUri'] = ['SedDocument:[0]', 'Simulation:[0]', 'numberOfSteps']

    signal = signal_name_to_signal['Time_step']
    signal['value'] = 0
    signal['bind']['min'] = 0
    signal['bind']['max'] = {
        'sedmlUri': ['SedDocument:[0]', 'Simulation:[0]', 'numberOfSteps'],
    }
    signal['bind']['step'] = 1

    # assign values to the Vega data sets
    data_set_name_to_index = {data_set['name']: i_data_set for i_data_set, data_set in enumerate(vega['data'])}
    vega['data'][data_set_name_to_index['glyphsData']]['values'] = glyphs
    vega['data'][data_set_name_to_index['arcsData']]['values'] = arcs
    vega['data'][data_set_name_to_index['glyphsValues']] = copy.copy(glyph_values_data_set)
    vega['data'][data_set_name_to_index['glyphsValues']]['name'] = 'glyphsValues'

    # export the Vega visualization
    with open(vega_output_filename, 'w') as file:
        json.dump(vega, file, indent=indent)



""" Data model for reports

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum


__all__ = [
    'VariableResults',
    'DataGeneratorResults',
    'DataSetResults',
    'ReportResults',
    'ReportFormat',
    'SedDocumentResults',
]


class VariableResults(dict):
    """ Dictionary that maps the ids of variables (:obj:`Variable`) to their results (:obj:`numpy.ndarray`)

    * Keys (:obj:`str`): ids of variables (:obj:`Variable`)
    * Values (:obj:`numpy.ndarray`): result of each variable

        * Steady-state tasks of non-spatial models: results should be arrays of shape ``(1,)``
        * One-step tasks of non-spatial models: results should be arrays of shape ``(2,)``
        * Uniform time course tasks of non-spatial models: results should be arrays of shape ``(number_of_steps + 1,)``
    """
    pass


class DataGeneratorResults(dict):
    """ Dictionary that maps the ids of data generators (:obj:`DataGenerator`) to their results (:obj:`numpy.ndarray`)

    * Keys (:obj:`str`): ids of data generators (:obj:`DataGenerator`)
    * Values (:obj:`numpy.ndarray`): result of each variable

        * Steady-state tasks of non-spatial models: results should be arrays of shape ``(1,)``
        * One-step tasks of non-spatial models: results should be arrays of shape ``(2,)``
        * Uniform time course tasks of non-spatial models: results should be arrays of shape ``(number_of_steps + 1,)``
    """
    pass


class DataSetResults(dict):
    """ Dictionary that maps the ids of data sets to their results (:obj:`numpy.ndarray`)

    * Keys (:obj:`str`): ids of data sets
    * Values (:obj:`numpy.ndarray`): result of each data set

        * Steady-state tasks of non-spatial models: results should be arrays of shape ``(number of data sets, 1)``
        * One-step tasks of non-spatial models: results should be arrays of shape ``(number of data sets, 2)``
        * Uniform time course tasks of non-spatial models: results should be arrays of shape ``(number_of_steps + 1,)``
    """
    pass


class ReportResults(dict):
    """ Dictionary that maps the ids of reports (e.g., :obj:`Report`) to their results (:obj:`DataSetResults`)

    * Keys (:obj:`str`): ids of reports (e.g., :obj:`Report`)
    * Values (:obj:`DataSetResults`): result of each report
    """
    pass


class ReportFormat(str, enum.Enum):
    """ Format of a report """
    csv = 'csv'
    h5 = 'h5'
    hdf = 'h5'
    hdf5 = 'h5'
    tsv = 'tsv'
    xlsx = 'xlsx'


class SedDocumentResults(dict):
    """ Dictionary that maps the locations of SED-ML documents (e.g., :obj:`SedDocument`) to their results (:obj:`ReportResults`)

    * Keys (:obj:`str`): locations of SED documents (e.g., :obj:`SedDocument`)
    * Values (:obj:`ReportResults`): result of each document
    """
    pass



""" Utilities for working with CellML models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (  # noqa: F401
    SedDocument, ModelAttributeChange, Variable, Symbol,
    Simulation, OneStepSimulation, UniformTimeCourseSimulation,
    Algorithm, Task,
    )
from ...utils.core import flatten_nested_list_of_strings
from .validation import validate_model
import libcellml  # noqa: F401
import lxml  # noqa: F401
import os
import types  # noqa: F401

__all__ = ['get_parameters_variables_outputs_for_simulation']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:cellml``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    # check model file exists
    if not isinstance(model_filename, str):
        raise ValueError('`{}` is not a path to a model file.'.format(model_filename))

    if not os.path.isfile(model_filename):
        raise FileNotFoundError('Model file `{}` does not exist.'.format(model_filename))

    errors, _, (model, root) = validate_model(model_filename, resolve_imports=False)
    if errors:
        raise ValueError('Model file `{}` is not a valid CellML file.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))

    if simulation_type not in [OneStepSimulation, UniformTimeCourseSimulation]:
        raise NotImplementedError('`simulation_type` must be `OneStepSimulation` or `UniformTimeCourseSimulation`')

    default_ns = root.nsmap.get(None, '')
    if (
        default_ns.startswith('http://www.cellml.org/cellml/1.0')
        or default_ns.startswith('http://www.cellml.org/cellml/1.1')
    ):
        return get_parameters_variables_for_simulation_version_1(model, root, simulation_type, algorithm_kisao_id=algorithm_kisao_id,
                                                                 native_ids=native_ids, native_data_types=native_data_types)

    else:
        return get_parameters_variables_for_simulation_version_2(model, root, simulation_type, algorithm_kisao_id=algorithm_kisao_id,
                                                                 native_ids=native_ids, native_data_types=native_data_types)


def get_parameters_variables_for_simulation_version_1(model, xml_root, simulation_type,
                                                      algorithm_kisao_id=None, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model (:obj:`None`): model
        xml_root (:obj:`lxml.etree._Element`): element tree for model
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    params = []
    if simulation_type == UniformTimeCourseSimulation:
        sim = UniformTimeCourseSimulation(
            id='simulation',
            initial_time=0.,
            output_start_time=0.,
            output_end_time=1.,
            number_of_steps=0.,
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id or 'KISAO_0000019',
            ),
        )
    else:
        sim = OneStepSimulation(
            id='simulation',
            step=1.,
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id or 'KISAO_0000019',
            ),
        )

    vars = []

    namespaces = {
        'cellml': xml_root.nsmap.get(None, '')
    }

    for component in xml_root.xpath('/cellml:model/cellml:component', namespaces=namespaces):
        component_name = component.attrib['name']
        for variable in component.xpath('cellml:variable', namespaces=namespaces):
            variable_name = variable.attrib['name']
            initial_value = variable.attrib.get('initial_value', None)
            if initial_value is not None:
                params.append(ModelAttributeChange(
                    id='{}.{}'.format(component_name, variable_name) if native_ids else 'initial_value_component_{}_variable_{}'.format(
                        component_name, variable_name),
                    name=None if native_ids else 'Initial value of variable "{}" of component "{}"'.format(
                        variable_name, component_name),
                    target="/cellml:model/cellml:component[@name='{}']/cellml:variable[@name='{}']/@initial_value".format(
                        component_name, variable_name),
                    target_namespaces=namespaces,
                    new_value=float(initial_value) if native_data_types else initial_value,
                ))

            vars.append(Variable(
                id='{}.{}'.format(component_name, variable_name) if native_ids else 'value_component_{}_variable_{}'.format(
                    component_name, variable_name),
                name=None if native_ids else 'Value of variable "{}" of component "{}"'.format(
                    variable_name, component_name),
                target="/cellml:model/cellml:component[@name='{}']/cellml:variable[@name='{}']".format(
                    component_name, variable_name),
                target_namespaces=namespaces,
            ))
    return params, [sim], vars, []


def get_parameters_variables_for_simulation_version_2(model, xml_root, simulation_type,
                                                      algorithm_kisao_id=None, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model (:obj:`libcellml.model.Model`): model
        xml_root (:obj:`lxml.etree._Element`): element tree for model
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`, optional): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
        :obj:`list` of :obj:`Plot`: possible plots of the results of a simulation of the model
    """
    namespaces = {
        'cellml': xml_root.nsmap.get(None, '')
    }

    params = []
    if simulation_type == UniformTimeCourseSimulation:
        sim = UniformTimeCourseSimulation(
            id='simulation',
            initial_time=0.,
            output_start_time=0.,
            output_end_time=1.,
            number_of_steps=0.,
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id or 'KISAO_0000019',
            ),
        )
    else:
        sim = OneStepSimulation(
            id='simulation',
            step=1.,
            algorithm=Algorithm(
                kisao_id=algorithm_kisao_id or 'KISAO_0000019',
            ),
        )
    vars = []

    for i_component in range(model.componentCount()):
        component = model.component(i_component)
        component_name = component.name()
        for i_variable in range(component.variableCount()):
            variable = component.variable(i_variable)

            variable_name = variable.name()
            initial_value = variable.initialValue()
            if initial_value:
                params.append(ModelAttributeChange(
                    id='{}.{}'.format(component_name, variable_name) if native_ids else 'initial_value_component_{}_variable_{}'.format(
                        component_name, variable_name),
                    name=None if native_ids else 'Initial value of variable "{}" of component "{}"'.format(
                        variable_name, component_name),
                    target="/cellml:model/cellml:component[@name='{}']/cellml:variable[@name='{}']/@initial_value".format(
                        component_name, variable_name),
                    target_namespaces=namespaces,
                    new_value=float(initial_value) if native_data_types else initial_value,
                ))

            vars.append(Variable(
                id='{}.{}'.format(component_name, variable_name) if native_ids else 'value_component_{}_variable_{}'.format(
                    component_name, variable_name),
                name=None if native_ids else 'Value of variable "{}" of component "{}"'.format(
                    variable_name, component_name),
                target="/cellml:model/cellml:component[@name='{}']/cellml:variable[@name='{}']".format(
                    component_name, variable_name),
                target_namespaces=namespaces,
            ))
    return params, [sim], vars, []



""" Data model for SED

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import evalidate
import math
import mpmath
import numpy
import numpy.random

__all__ = [
    'log',
    'piecewise',
    'MATHEMATICAL_FUNCTIONS',
    'RESERVED_MATHEMATICAL_SYMBOLS',
    'AGGREGATE_MATH_FUNCTIONS',
    'VALID_MATH_EXPRESSION_NODES',
    'compile_math',
    'eval_math',
]


def log(*args):
    """ Evaluate a logarithm

    Args:
        *args (:obj:`list` of :obj:`float`): value optional proceeded by a base; otherwise the logarithm
            is calculated in base 10

    Returns:
        :obj:`float`
    """
    value = args[-1]
    if len(args) > 1:
        base = args[0]
    else:
        base = 10.

    return math.log(value, base)


def piecewise(*args):
    """ Evaluate a MathML piecewise function

    Args:
        *args (:obj:`list` of :obj:`float`): pairs of value and conditions followed by a default value

    Returns:
        :obj:`float`
    """
    if len(args) % 2 == 0:
        pieces = args
        otherwise = math.nan

    else:
        pieces = args[0:-1]
        otherwise = args[-1]

    for i_piece in range(0, len(pieces), 2):
        value = pieces[i_piece]
        condition = pieces[i_piece + 1]
        if condition:
            return value

    return otherwise


MATHEMATICAL_FUNCTIONS = {
    'root': lambda x, n: x**(1 / float(n)),
    'abs': abs,
    'exp': math.exp,
    'ln': math.log,
    'log': log,
    'floor': math.floor,
    'ceiling': math.ceil,
    'factorial': math.factorial,
    'sin': math.sin,
    'cos': math.cos,
    'tan': math.tan,
    'sec': mpmath.sec,
    'csc': mpmath.csc,
    'cot': mpmath.cot,
    'sinh': math.sinh,
    'cosh': math.cosh,
    'tanh': math.tanh,
    'sech': mpmath.sech,
    'csch': mpmath.csch,
    'coth': mpmath.coth,
    'arcsin': math.asin,
    'arccos': math.acos,
    'arctan': math.atan,
    'arcsec': mpmath.asec,
    'arccsc': mpmath.acsc,
    'arccot': mpmath.acot,
    'arcsinh': math.asinh,
    'arccosh': math.acosh,
    'arctanh': math.atanh,
    'arcsech': mpmath.asech,
    'arccsch': mpmath.acsch,
    'arccoth': mpmath.acoth,
    'min': numpy.min,
    'max': numpy.max,
    'sum': numpy.sum,
    'product': numpy.product,
    'count': len,
    'mean': numpy.mean,
    'stdev': numpy.std,
    'variance': numpy.var,
    'uniform': numpy.random.uniform,
    'normal': numpy.random.normal,
    'lognormal': numpy.random.lognormal,
    'poisson': numpy.random.poisson,
    'gamma': numpy.random.gamma,
    'piecewise': piecewise,
}

RESERVED_MATHEMATICAL_SYMBOLS = {
    'true': True,
    'false': False,
    'notanumber': math.nan,
    'pi': math.pi,
    'infinity': math.inf,
    'exponentiale': math.e,
}

AGGREGATE_MATH_FUNCTIONS = (
    'min',
    'max',
    'sum',
    'product',
    'count',
    'mean',
    'stdev',
    'variance',
)


VALID_MATH_EXPRESSION_NODES = [
    'Eq',
    'NotEq',
    'Gt',
    'Lt',
    'GtE',
    'LtE',
    'Sub',
    'Mult',
    'Div',
    'Pow',
    'And',
    'Or',
    'Not',
    'BitAnd',
    'BitOr',
    'BitXor',
    'Call',
    'Constant',
]


def compile_math(math):
    """ Compile a mathematical expression

    Args:
        math (:obj:`str`): mathematical expression

    Returns:
        :obj:`_ast.Expression`: compiled expression
    """
    if isinstance(math, str):
        math = (
            math
            .replace('&&', 'and')
            .replace('||', 'or')
        )

    math_node = evalidate.evalidate(math,
                                    addnodes=VALID_MATH_EXPRESSION_NODES,
                                    funcs=MATHEMATICAL_FUNCTIONS.keys())
    compiled_math = compile(math_node, '<math>', 'eval')
    return compiled_math


def eval_math(math, compiled_math, workspace):
    """ Compile a mathematical expression

    Args:
        math (:obj:`str`): mathematical expression
        compiled_math (:obj:`_ast.Expression`): compiled expression
        workspace (:obj:`dict`): values to use for the symbols in the expression

    Returns:
        :obj:`object`: result of the expression

    Raises:
        :obj:`ValueError`: if the expression could not be evaluated
    """
    invalid_symbols = set(RESERVED_MATHEMATICAL_SYMBOLS.keys()).intersection(set(workspace.keys()))
    if invalid_symbols:
        raise ValueError('Variables for mathematical expressions cannot have ids equal to the following reserved symbols:\n  - {}'.format(
            '\n  - '.join('`' + symbol + '`' for symbol in sorted(invalid_symbols))))

    try:
        return eval(compiled_math, MATHEMATICAL_FUNCTIONS, dict(**RESERVED_MATHEMATICAL_SYMBOLS, **workspace))
    except Exception as exception:
        raise ValueError('Expression `{}` could not be evaluated:\n\n  {}\n\n  workspace:\n    {}'.format(
            math, str(exception), '\n    '.join('{}: {}'.format(key, value) for key, value in workspace.items())))



""" Utility for converting metabolic maps from Escher to Vega format

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-02-06
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...utils import cubic_bezier_point, cubic_bezier_angle
import bezier
import copy
import json
import math
import numpy
import os
import pkg_resources

__all__ = ['escher_to_vega', 'read_escher_map_config']


def escher_to_vega(reaction_fluxes_data_set, escher_filename, vega_filename,
                   max_width_height=800, legend_padding=20, legend_width=40, signal_height=20,
                   arrow_head_gap=16., indent=2, metabolic_map_data_file=None):
    """ Convert a metabolic pathway map from Escher format to Vega format.

    Args:
        reaction_fluxes_data_set (:obj:`dict`): Vega data set with each predicted reaction flux. The dictionary should
            have one of the following schemas

            * Contain a key ``values`` whose value is a list of the predicted flux values. Each predicted flux value
              should be a dictionary with two keys

                * ``label``: value should be the BiGG identifier of the corresponding reaction
                * ``values``: array with a single value equal to the predicted flux

            * Contain a key ``url`` whose value is a URL where the predicted flux values can be retrieved in JSON
              format. The data should be in the same format outlined above.
            * Contain a key ``sedmlUri`` whose value is an array which contains two elements that indicate

                * The location of the SED-ML file (within its parent COMBINE/OMEX archive) which will produce
                  the flux predictions for the visualization
                * The id of the SED-ML report which will contain the flux predictions for the visualization

        escher_filename (:obj:`str`): path to the visualization in Escher format
        vega_filename (:obj:`str`): path to save the visualization in Vega format
        max_width_height (:obj:`int`, optional): maximum height/width of the metabolic map in pixels
        legend_padding (:obj:`int`, optional): horizontal spacing between the metabolic map and legend in pixels
        legend_width (:obj:`int`, optional): legend width in pixels, including the width of the title
        signal_height (:obj:`int`, optional): height of signals in pixels
        arrow_head_gap (:obj:`float`, optional): spacing between arrow heads and nodes
        indent (:obj:`int` or :obj:`None`, optional): indentation
        metabolic_map_data_file (:obj:`dict` or :obj:`None`, optional): If :obj:`None`, store the data for
            the metabolic map (e.g., coordinates of metabolites, reactions) inline within the generated Vega
            file. If not :obj:`None`, store the data for the metabolic map in a separate file to be loaded
            via a URL. In this case, the argument must be a dictionary which two keys, ``filename`` and ``url``
            which indicate where the data should be saved and where this file will be hosted.
    """
    # read the Escher version of the map
    escher = read_escher_map_config(escher_filename)

    # determine the extent (min/max x/y) of the map
    min_x = math.inf
    min_y = math.inf
    max_x = -math.inf
    max_y = -math.inf

    nodes = escher[1]['nodes']
    for node in nodes.values():
        min_x = min(min_x, node['x'])
        min_y = min(min_y, node['y'])
        max_x = max(max_x, node['x'])
        max_y = max(max_y, node['y'])

        if 'label_x' in node:
            min_x = min(min_x, node['label_x'])
            min_y = min(min_y, node['label_y'])
            max_x = max(max_x, node['label_x'])
            max_y = max(max_y, node['label_y'])

    for reaction in escher[1]['reactions'].values():
        if 'label_x' in reaction:
            min_x = min(min_x, reaction['label_x'])
            min_y = min(min_y, reaction['label_y'])
            max_x = max(max_x, reaction['label_x'])
            max_y = max(max_y, reaction['label_y'])

    escher_width = max_x - min_x
    escher_height = max_y - min_y

    if escher_width > escher_height:
        width = max_width_height
        coordinate_scale = width / escher_width
        height = escher_height * coordinate_scale
    else:
        height = max_width_height
        coordinate_scale = height / escher_height
        width = escher_width * coordinate_scale

    # convert metabolite circle nodes to Vega
    metabolites = []
    for id, node in nodes.items():
        if node["node_type"] == "metabolite":
            metabolite = {}

            metabolite["id"] = id
            metabolite["biggId"] = node['bigg_id']
            metabolite["name"] = node['name']

            metabolite["x"] = (node["x"] - min_x) * coordinate_scale
            metabolite["y"] = (node["y"] - min_y) * coordinate_scale

            if node.get('node_is_primary', False):
                metabolite['size'] = 10 ** 2 * coordinate_scale
            else:
                metabolite['size'] = 20 ** 2 * coordinate_scale

            if "label_x" in node:
                metabolite["labelX"] = (node["label_x"] - min_x) * coordinate_scale
                metabolite["labelY"] = (node["label_y"] - min_y) * coordinate_scale

            reaction_ids = set()
            for rxn_id, reaction in escher[1]['reactions'].items():
                for segment in reaction["segments"].values():
                    if segment["from_node_id"] == id or segment["to_node_id"] == id:
                        reaction_ids.add(rxn_id)
                        break
            metabolite['reactionIds'] = list(reaction_ids)
            metabolite['metaboliteIds'] = []

            related_metabolite_ids = set()
            for reaction_id in reaction_ids:
                reaction = escher[1]['reactions'][reaction_id]

                for segment in reaction["segments"].values():
                    from_node = nodes[segment["from_node_id"]]
                    to_node = nodes[segment["to_node_id"]]

                    if from_node["node_type"] == "metabolite":
                        related_metabolite_ids.add(segment["from_node_id"])
                    if to_node['node_type'] == 'metabolite':
                        related_metabolite_ids.add(segment["to_node_id"])
            metabolite['relatedMetaboliteIds'] = list(related_metabolite_ids)

            metabolites.append(metabolite)

    # convert reaction paths to Vega
    reaction_segment_coordinates = []
    reaction_arrow_head_coordinates = []
    reaction_labels = []
    for id, reaction in escher[1]['reactions'].items():
        metabolite_ids = set()
        for rxn_id, other_reaction in escher[1]['reactions'].items():
            if other_reaction['bigg_id'] == reaction['bigg_id']:
                for segment in reaction["segments"].values():
                    from_node = nodes[segment["from_node_id"]]
                    to_node = nodes[segment["to_node_id"]]

                    if from_node["node_type"] == "metabolite":
                        metabolite_ids.add(segment["from_node_id"])
                    if to_node['node_type'] == 'metabolite':
                        metabolite_ids.add(segment["to_node_id"])
        metabolite_ids = list(metabolite_ids)

        reaction_labels.append({
            "id": id,
            "biggId": reaction["bigg_id"],
            "x": (reaction["label_x"] - min_x) * coordinate_scale,
            "y": (reaction["label_y"] - min_y) * coordinate_scale,
            'reactionIds': [],
            'metaboliteIds': metabolite_ids,
            'relatedMetaboliteIds': [],
        })

        for segment in reaction["segments"].values():
            from_node = nodes[segment["from_node_id"]]
            to_node = nodes[segment["to_node_id"]]

            if from_node['node_type'] == 'metabolite':
                if segment['b1']:
                    points = [
                        [from_node['x'], from_node['y']],
                        [segment['b1']['x'], segment['b1']['y']],
                        [segment['b2']['x'], segment['b2']['y']],
                        [to_node['x'], to_node['y']],
                    ]
                    segment_len = bezier.Curve(numpy.array(points).transpose(), degree=3).length
                    t = min(1., arrow_head_gap / segment_len)
                    x0, y0 = cubic_bezier_point(points, t)
                    angle = cubic_bezier_angle(points, t) - 90.
                else:
                    segment_len = math.sqrt((from_node['x'] - to_node['x']) ** 2 + (from_node['y'] - to_node['y']) ** 2)
                    angle = math.atan2(from_node['y'] - to_node['y'], from_node['x'] - to_node['x'])
                    t = min(1., arrow_head_gap / segment_len)
                    x0 = from_node["x"] * (1 - t) + to_node["x"] * t
                    y0 = from_node["y"] * (1 - t) + to_node["y"] * t

                reaction_arrow_head_coordinates.append({
                    'id': id,
                    'biggId': reaction['bigg_id'],
                    'start': True,
                    'x': (x0 - min_x) * coordinate_scale,
                    'y': (y0 - min_y) * coordinate_scale,
                    'angle': angle,
                    'reactionIds': [],
                    'metaboliteIds': metabolite_ids,
                    'relatedMetaboliteIds': [],
                })
            else:
                x0 = from_node["x"]
                y0 = from_node["y"]

            if to_node['node_type'] == 'metabolite':
                if segment['b1']:
                    points = [
                        [from_node['x'], from_node['y']],
                        [segment['b1']['x'], segment['b1']['y']],
                        [segment['b2']['x'], segment['b2']['y']],
                        [to_node['x'], to_node['y']],
                    ]
                    segment_len = bezier.Curve(numpy.array(points).transpose(), degree=3).length
                    t = max(0., 1. - arrow_head_gap / segment_len)
                    x3, y3 = cubic_bezier_point(points, t)
                    angle = cubic_bezier_angle(points, t) + 90.
                else:
                    segment_len = math.sqrt((from_node['x'] - to_node['x']) ** 2 + (from_node['y'] - to_node['y']) ** 2)
                    angle = math.atan2(to_node['y'] - from_node['y'], to_node['x'] - from_node['x'])
                    t = max(0., 1. - arrow_head_gap / segment_len)
                    x3 = from_node["x"] * (1 - t) + to_node["x"] * t
                    y3 = from_node["y"] * (1 - t) + to_node["y"] * t

                reaction_arrow_head_coordinates.append({
                    'id': id,
                    'biggId': reaction['bigg_id'],
                    'start': False,
                    'x': (x3 - min_x) * coordinate_scale,
                    'y': (y3 - min_y) * coordinate_scale,
                    'angle': angle,
                    'reactionIds': [],
                    'metaboliteIds': metabolite_ids,
                    'relatedMetaboliteIds': [],
                })
            else:
                x3 = to_node["x"]
                y3 = to_node["y"]

            if segment["b1"]:
                reaction_segment_coordinates.append({
                    "id": id,
                    "biggId": reaction["bigg_id"],
                    "name": reaction["name"],
                    'type': 'curve',
                    'x0': (x0 - min_x) * coordinate_scale,
                    'y0': (y0 - min_y) * coordinate_scale,
                    'x1': (segment["b1"]["x"] - min_x) * coordinate_scale,
                    'y1': (segment["b1"]["y"] - min_y) * coordinate_scale,
                    'x2': (segment["b2"]["x"] - min_x) * coordinate_scale,
                    'y2': (segment["b2"]["y"] - min_y) * coordinate_scale,
                    'x3': (x3 - min_x) * coordinate_scale,
                    'y3': (y3 - min_y) * coordinate_scale,
                    'reactionIds': [],
                    'metaboliteIds': metabolite_ids,
                    'relatedMetaboliteIds': [],
                })
            else:
                reaction_segment_coordinates.append({
                    "id": id,
                    "biggId": reaction["bigg_id"],
                    "name": reaction["name"],
                    'type': 'line',
                    'x0': (x0 - min_x) * coordinate_scale,
                    'y0': (y0 - min_y) * coordinate_scale,
                    'x1': (x3 - min_x) * coordinate_scale,
                    'y1': (y3 - min_y) * coordinate_scale,
                    'reactionIds': [],
                    'metaboliteIds': metabolite_ids,
                    'relatedMetaboliteIds': [],
                })

    # insert metabolite circles and reaction paths into Vega template
    template_filename = pkg_resources.resource_filename(
        'biosimulators_utils',
        os.path.join('viz', 'vega', 'escher', 'template.json'))
    with open(template_filename, 'r') as file:
        vega = json.load(file)

    vega['width'] = width + legend_padding + legend_width
    vega['height'] = height + signal_height

    map_width_signal = next(signal for signal in vega['signals'] if signal['name'] == 'mapMaxX')
    map_width_signal['value'] = width

    map_height_signal = next(signal for signal in vega['signals'] if signal['name'] == 'mapMaxY')
    map_height_signal['value'] = height

    legend_width_signal = next(signal for signal in vega['signals'] if signal['name'] == 'legendWidth')
    legend_width_signal['value'] = legend_width

    legend_padding_signal = next(signal for signal in vega['signals'] if signal['name'] == 'legendPadding')
    legend_padding_signal['value'] = legend_padding

    signal_height_signal = next(signal for signal in vega['signals'] if signal['name'] == 'signalHeight')
    signal_height_signal['value'] = signal_height

    signal_padding_signal = next(signal for signal in vega['signals'] if signal['name'] == 'signalPadding')
    signal_padding_signal['value'] = 0

    metabolite_stroke_width_signal = next(signal for signal in vega['signals'] if signal['name'] == 'metaboliteStrokeWidthData')
    metabolite_stroke_width_signal['value'] = 2 * coordinate_scale

    reaction_stroke_width_signal = next(signal for signal in vega['signals'] if signal['name'] == 'reactionStrokeWidthData')
    reaction_stroke_width_signal['value'] = 18 * coordinate_scale

    arrow_head_stroke_width_signal = next(signal for signal in vega['signals'] if signal['name'] == 'arrowHeadStrokeWidthData')
    arrow_head_stroke_width_signal['value'] = 1 * coordinate_scale

    metabolic_map_data = {
        "metabolites": metabolites,
        "reactionSegmentCoordinates": reaction_segment_coordinates,
        "reactionArrowHeadCoordinates": reaction_arrow_head_coordinates,
        "reactionLabels": reaction_labels,
    }
    if metabolic_map_data_file:
        with open(metabolic_map_data_file['filename'], 'w') as file:
            json.dump(metabolic_map_data, file, indent=indent)

        for key in metabolic_map_data.keys():
            vega_data_set = next(data for data in vega['data'] if data['name'] == key)
            del vega_data_set['values']
            vega_data_set['url'] = metabolic_map_data_file['url']
            vega_data_set['format'] = {
                'type': 'json',
                'property': key
            }
    else:
        for key, values in metabolic_map_data.items():
            vega_data_set = next(data for data in vega['data'] if data['name'] == key)
            vega_data_set['values'] = values

    i_reaction_fluxes_data_set = next(i_data for i_data, data in enumerate(vega['data']) if data['name'] == 'reactionFluxes')
    vega['data'][i_reaction_fluxes_data_set] = copy.copy(reaction_fluxes_data_set)
    vega['data'][i_reaction_fluxes_data_set]['name'] = 'reactionFluxes'

    # save Vega-formatted map
    with open(vega_filename, 'w') as file:
        json.dump(vega, file, indent=indent)


def read_escher_map_config(filename):
    """ Read the configuration of an Escher map from a file

    Args:
        filename (:obj:`str`): path to a configuration of an Escher map

    Returns:
        :obj:`list`: configuration of an Escher map
    """
    with open(filename, 'r') as file:
        return json.load(file)



""" Utilities for reading the specifications of a simulator

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..sedml.data_model import (SedDocument, Task, ModelLanguagePattern, ModelLanguageEdamId,  # noqa: F401
                                SteadyStateSimulation, UniformTimeCourseSimulation, Algorithm, AlgorithmParameterChange)
from .data_model import SoftwareInterface
from kisao import Kisao
from kisao.data_model import AlgorithmSubstitutionPolicy
from kisao.utils import get_substitutable_algorithms_for_policy
import json
import natsort
import re
import requests

__all__ = [
    'BIOSIMULATORS_API_ENDPOINT',
    'get_simulator_specs',
    'does_simulator_have_capabilities_to_execute_sed_document',
    'does_simulator_have_capabilities_to_execute_sed_task',
    'does_algorithm_implementation_have_capabilities_to_execute_sed_task',
    'does_algorithm_implementation_have_capabilities_to_execute_sed_model_language',
    'does_algorithm_implementation_have_capabilities_to_execute_parameter',
    'gen_algorithms_from_specs',
]


BIOSIMULATORS_API_ENDPOINT = 'https://api.biosimulators.org/simulators'


def get_simulator_specs(id, version='latest'):
    """ Get the specifications of a simulation tool from the BioSimulators registry

    Args:
        id (:obj:`str`): id
        version (:obj:`str`, optional): version

    Returns:
        :obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Simulator``: specifications
    """
    if version == 'latest':
        url = BIOSIMULATORS_API_ENDPOINT + '/' + id
        response = requests.get(url)
        response.raise_for_status()
        version_specs = natsort.natsorted(response.json(), key=lambda version_spec: version_spec['version'])
        return version_specs[-1]
    else:
        url = BIOSIMULATORS_API_ENDPOINT + '/' + id + '/' + version
        response = requests.get(url)
        response.raise_for_status()
        return response.json()


def does_simulator_have_capabilities_to_execute_sed_document(sed_doc, simulator_specs,
                                                             alg_substitution_policy=AlgorithmSubstitutionPolicy.SAME_VARIABLES):
    """ Determine if a simulator has the capabilities to execute a SED document

    Args:
        sed_doc (:obj:`SedDocument`): SED document
        simulator_specs (:obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Simulator``):
            specifications of a simulation tool
        alg_substitution_policy (:obj:`AlgorithmSubstitutionPolicy`, optional): algorithm substitution policy

    Returns:
        :obj:`bool`: whether the simulator has the capabilities to execute the SED document
    """
    for task in sed_doc.tasks:
        if not does_simulator_have_capabilities_to_execute_sed_task(task, simulator_specs, alg_substitution_policy=alg_substitution_policy):
            return False
    return True


def does_simulator_have_capabilities_to_execute_sed_task(task, simulator_specs,
                                                         alg_substitution_policy=AlgorithmSubstitutionPolicy.SAME_VARIABLES):
    """ Determine if a simulator has the capabilities to execute a SED task

    Args:
        task (:obj:`Task`): SED task
        simulator_specs (:obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Simulator``):
            specifications of a simulation tool
        alg_substitution_policy (:obj:`AlgorithmSubstitutionPolicy`, optional): algorithm substitution policy

    Returns:
        :obj:`bool`: whether the simulator has the capabilities to execute the SED task
    """
    if isinstance(task, Task):
        if not isinstance(task.simulation, (SteadyStateSimulation, UniformTimeCourseSimulation)):
            return False

        for alg_specs in simulator_specs['algorithms']:
            if does_algorithm_implementation_have_capabilities_to_execute_sed_task(task, alg_specs,
                                                                                   alg_substitution_policy=alg_substitution_policy):
                return True

        return False

    else:
        for sub_task in task.sub_tasks:
            if not does_simulator_have_capabilities_to_execute_sed_task(sub_task.task, simulator_specs,
                                                                        alg_substitution_policy=alg_substitution_policy):
                return False
        return True


def does_algorithm_implementation_have_capabilities_to_execute_sed_task(task, algorithm_specs,
                                                                        alg_substitution_policy=AlgorithmSubstitutionPolicy.SAME_VARIABLES):
    """ Determine if an implementation of an algorithm has the capabilities to execute a SED task

    Args:
        task (:obj:`Task`): SED task
        algorithm_specs (:obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Algorithm``):
            specifications of the implementation of an algorithm
        alg_substitution_policy (:obj:`AlgorithmSubstitutionPolicy`, optional): algorithm substitution policy

    Returns:
        :obj:`bool`: whether the implementation of the algorithm has the capabilities to execute the SED task
    """
    model = task.model
    simulation = task.simulation
    algorithm = simulation.algorithm

    kisao = Kisao()
    alg_term = kisao.get_term(algorithm.kisao_id)
    alt_algs = get_substitutable_algorithms_for_policy(alg_term, substitution_policy=alg_substitution_policy)
    alt_alg_ids = kisao.get_term_ids(alt_algs)

    if algorithm_specs['kisaoId']['id'] in alt_alg_ids:
        # check if the implementation supports the model language
        if not does_algorithm_implementation_have_capabilities_to_execute_sed_model_language(model.language, algorithm_specs):
            return False

        # check if implementation supports the parameters of the algorithm
        supports_parameters = True
        if algorithm_specs['kisaoId']['id'] == algorithm.kisao_id:
            for change in algorithm.changes:
                if not does_algorithm_implementation_have_capabilities_to_execute_parameter(change.kisao_id, algorithm_specs):
                    supports_parameters = False
                    break

        if not supports_parameters:
            return False

        return True

    return False


def does_algorithm_implementation_have_capabilities_to_execute_sed_model_language(model_language, algorithm_specs):
    """ Determine if an implementation of an algorithm has the capabilities to execute a model langugae

    Args:
        model_language (:obj:`str`): SED URN for model language
        algorithm_specs (:obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Algorithm``):
            specifications of the implementation of an algorithm

    Returns:
        :obj:`bool`: whether the implementation of the algorithm has the capabilities to execute the SED model language
    """
    model_language_edam_id = None
    for model_language_pattern in ModelLanguagePattern.__members__.values():
        if model_language and re.match(model_language_pattern.value, model_language):
            model_language_edam_id = ModelLanguageEdamId[model_language_pattern.name].value
            break
    if not model_language_edam_id:
        return False

    for model_format_specs in algorithm_specs['modelFormats']:
        if model_format_specs['id'] == model_language_edam_id:
            return True
    return False


def does_algorithm_implementation_have_capabilities_to_execute_parameter(parameter_kisao_id, algorithm_specs):
    """ Determine if an implementation of an algorithm has the capabilities to execute a model langugae

    Args:
        parameter_kisao_id (:obj:`str`): KiSAO id for an algorithm parameter
        algorithm_specs (:obj:`dict` with schema ``https://api.biosimulators.org/openapi.json#/components/schemas/Algorithm``):
            specifications of the implementation of an algorithm

    Returns:
        :obj:`bool`: whether the implementation of the algorithm has the capabilities to execute the SED parameter
    """
    for parameter_specs in algorithm_specs['parameters']:
        if parameter_specs['kisaoId']['id'] == parameter_kisao_id:
            return True
    return False


def gen_algorithms_from_specs(specifications):
    """ Generate a list of algorithms and their parameters from the specifications of a simulator

    Args:
        specifications (:obj:`dict` or :obj:`str`): specifications or path to specifications

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`Algorithm`: dictionary that maps KiSAO ids to algorithms and their parameters
    """
    if isinstance(specifications, str):
        with open(specifications, 'rb') as file:
            specifications = json.load(file)

    algs = {}
    for alg_spec in specifications.get('algorithms', []):
        if SoftwareInterface.biosimulators_docker_image.value in alg_spec['availableSoftwareInterfaceTypes']:

            alg = Algorithm()
            alg.kisao_id = alg_spec.get('kisaoId', {}).get('id', None)
            algs[alg.kisao_id] = alg

            param_specs = alg_spec.get('parameters', None)
            if param_specs:
                for param_spec in param_specs:
                    if SoftwareInterface.biosimulators_docker_image.value in param_spec['availableSoftwareInterfaceTypes']:
                        alg.changes.append(AlgorithmParameterChange(
                            kisao_id=param_spec.get('kisaoId', {}).get('id', None),
                            new_value=param_spec.get('value', None),
                        ))

    return algs



""" Tool for managing Gurobi licenses through environment variables

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-09-1
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .core import LicenseManager
import functools
import os

try:
    import gurobipy  # noqa: F401
except ModuleNotFoundError:
    gurobipy = None


class GurobiLicenseManager(LicenseManager):
    """ Gurobi license manager """
    ENV_VAR_PREFIX = 'GRB'

    def __init__(self):
        if gurobipy is not None:
            self.__gurobipy_Model = gurobipy.Model
        else:
            self.__gurobipy_Model = None

    def get_keys_from_env_vars(self):
        """ Get the license keys for the software package from environment variables

        Returns:
            :obj:`dict`: environment license variables for a software package
        """
        keys = super(GurobiLicenseManager, self).get_keys_from_env_vars()
        license_id = keys.get('LICENSEID', None)
        if license_id is not None:
            keys['LICENSEID'] = int(float(license_id))
        return keys

    def is_package_available(self):
        """ Determine whether the package is installed

        Returns:
            :obj:`bool`: whether the package is installed
        """
        return (
            gurobipy is not None and (
                self.get_keys_from_env_vars()
                or (
                    os.path.isfile(os.path.expanduser(os.path.join('~', 'gurobi.lic')))
                    or os.path.isfile(os.path.join('/opt', 'gurobi', 'gurobi.lic'))
                )
            )
        )

    def start(self):
        """ Initialize usage of the software package with license keys """
        if self.is_package_available():
            keys = self.get_keys_from_env_vars()

            if keys:
                env = gurobipy.Env(params=keys)
                gurobipy.Model = functools.partial(self.__gurobipy_Model, env=env)  # to make OptLang use the Gurobi environment

    def end(self):
        """ Terminate usage of the software package with license keys """
        if self.is_package_available():
            gurobipy.Model = self.__gurobipy_Model

    def save_keys_to_license_file(self, filename=os.path.expanduser(os.path.join('~', 'gurobi.lic'))):
        """ Save license keys to a file

        Args:
            filename (:obj:`str`): path to save license
        """
        keys = self.get_keys_from_env_vars()
        if keys and not os.path.isfile(filename):
            with open(filename, 'w') as file:
                for key, val in keys.items():
                    file.write('{}={}\n'.format(key, val))



""" Data model for the execution status of COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
import enum
import os
import yaml

__all__ = [
    'Status',
    'Log',
    'CombineArchiveLog',
    'SedDocumentLog',
    'TaskLog',
    'OutputLog',
    'ReportLog',
    'Plot2DLog',
    'Plot3DLog',
    'StandardOutputErrorCapturerLevel',
]


class Status(str, enum.Enum):
    """ Status of COMBINE/OMEX archive or one of its components """
    QUEUED = 'QUEUED'

    RUNNING = 'RUNNING'

    SUCCEEDED = 'SUCCEEDED'

    SKIPPED = 'SKIPPED'
    # component will not / wasn't executed e.g., a plot won't be created because a
    # simulation doesn't have the ability to create plots

    FAILED = 'FAILED'


class Log(object):
    """ Log of a COMBINE/OMEX archive or one of its components

    Attributes:
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        parent (:obj:`Log`): execution status of parent COMBINE/OMEX archive
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None, parent=None, out_dir=None):
        """
        Args:
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            parent (:obj:`Log`, optional): execution status of parent COMBINE/OMEX archive
            out_dir (:obj:`str`, optional): directory to export status
        """
        self.status = status
        self.exception = exception
        self.skip_reason = skip_reason
        self.output = output
        self.duration = duration
        self.parent = parent
        self.out_dir = out_dir

    def finalize(self):
        """ Mark all unexecuted elements as skipped """
        if self.status == Status.QUEUED:
            self.status = Status.SKIPPED
        elif self.status == Status.RUNNING:
            self.status = Status.FAILED

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        value = {}

        value['status'] = self.status.value if self.status else None

        if self.exception:
            value['exception'] = {
                'type': self.exception.__class__.__name__,
                'message': str(self.exception),
            }
        else:
            value['exception'] = None

        if self.skip_reason:
            value['skipReason'] = {
                'type': self.skip_reason.__class__.__name__,
                'message': str(self.skip_reason),
            }
        else:
            value['skipReason'] = None

        value['output'] = self.output

        value['duration'] = self.duration

        return value

    def export(self):
        """ Write to a file """
        if self.out_dir:
            log_path = get_config().LOG_PATH
            if log_path:
                path = os.path.join(self.out_dir, log_path)
                if not os.path.isdir(self.out_dir):
                    os.makedirs(self.out_dir)
                with open(path, 'w') as file:
                    file.write(yaml.dump(self.to_json()))
        elif self.parent:
            self.parent.export()


class CombineArchiveLog(Log):
    """ Log of a COMBINE/OMEX archive

    Attributes:
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        sed_documents (:obj:`dict` of :obj:`str` to :obj:`SedDocumentLog`): execution status of each
            SED document in the archive
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None, sed_documents=None,
                 out_dir=None):
        """
        Args:
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            sed_documents (:obj:`dict` of :obj:`str` to :obj:`SedDocumentLog`, optional): execution status of each
                SED document in the archive
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(CombineArchiveLog, self).__init__(status=status, exception=exception,
                                                skip_reason=skip_reason, output=output,
                                                duration=duration, out_dir=out_dir)
        self.sed_documents = sed_documents

    def finalize(self):
        """ Mark all unexceuted elements as skipped """
        super(CombineArchiveLog, self).finalize()

        if self.sed_documents:
            for sed_document in self.sed_documents.values():
                sed_document.finalize()

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        value = super(CombineArchiveLog, self).to_json()
        value['sedDocuments'] = (
            [doc_log.to_json() for doc_log in self.sed_documents.values() if doc_log]
            if self.sed_documents is not None
            else None
        )
        return value


class SedDocumentLog(Log):
    """ Log of a SED document

    Attributes:
        location (:obj:`str`): location of the SED document
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        tasks (:obj:`dict` of :obj:`str` to :obj:`TaskLog`): execution status of each
            task
        outputs (:obj:`dict` of :obj:`str` to :obj:`OutputLog`): execution status of each
            output
        parent (:obj:`CombineArchiveLog`): execution status of parent COMBINE/OMEX archive
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, location=None, status=None, exception=None, skip_reason=None, output=None, duration=None,
                 tasks=None, outputs=None, parent=None, out_dir=None):
        """
        Args:
            location (:obj:`str`, optional): location of the SED document
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            tasks (:obj:`dict` of :obj:`str` to :obj:`TaskLog`, optional): execution status of each
                task
            outputs (:obj:`dict` of :obj:`str` to :obj:`OutputLog`, optional): execution status of each
                output
            parent (:obj:`CombineArchiveLog`, optional): execution status of parent COMBINE/OMEX archive
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(SedDocumentLog, self).__init__(status=status, exception=exception,
                                             skip_reason=skip_reason, output=output,
                                             duration=duration, parent=parent, out_dir=out_dir)
        self.location = location
        self.tasks = tasks
        self.outputs = outputs

    def finalize(self):
        """ Mark all unexceuted elements as skipped """
        super(SedDocumentLog, self).finalize()

        if self.tasks:
            for task in self.tasks.values():
                task.finalize()

        if self.outputs:
            for output in self.outputs.values():
                output.finalize()

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        value = super(SedDocumentLog, self).to_json()
        value['location'] = self.location
        value['tasks'] = (
            [task_log.to_json() for task_log in self.tasks.values() if task_log]
            if self.tasks is not None
            else None
        )
        value['outputs'] = (
            [output_log.to_json() for output_log in self.outputs.values() if output_log]
            if self.outputs is not None
            else None
        )
        return value


class TaskLog(Log):
    """ Log of a SED task

    Attributes:
        id (:obj:`str`): id of the task
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        algorithm (:obj:`str`): KiSAO id of the requested algorithm
        simulator_details (:obj:`dict`): additional simulator-specific information
        parent (:obj:`SedDocumentLog`): execution status of parent SED document
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None,
                 algorithm=None, simulator_details=None, parent=None, out_dir=None):
        """
        Args:
            id (:obj:`str`, optional): id of the task
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            algorithm (:obj:`str`, optional): KiSAO id of the executed algorithm
            simulator_details (:obj:`dict`, optional): additional simulator-specific information
            parent (:obj:`SedDocumentLog`): execution status of parent SED document
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(TaskLog, self).__init__(status=status, exception=exception,
                                      skip_reason=skip_reason, output=output,
                                      duration=duration, parent=parent, out_dir=out_dir)
        self.id = id
        self.algorithm = algorithm
        self.simulator_details = simulator_details

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        value = super(TaskLog, self).to_json()
        value['id'] = self.id
        value['algorithm'] = self.algorithm
        value['simulatorDetails'] = (
            [{'key': key, 'value': value} for key, value in self.simulator_details.items()]
            if self.simulator_details
            else None
        )
        return value


class OutputLog(Log):
    """ Log of a SED output

    Attributes:
        id (:obj:`str`): id of the output
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        parent (:obj:`SedDocumentLog`): execution status of parent SED document
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None,
                 parent=None, out_dir=None):
        """
        Args:
            id (:obj:`str`, optional): id of the report
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            parent (:obj:`SedDocumentLog`, optional): execution status of parent SED document
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(OutputLog, self).__init__(status=status, exception=exception, skip_reason=skip_reason,
                                        output=output, duration=duration, parent=parent, out_dir=out_dir)
        self.id = id

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        value = super(OutputLog, self).to_json()
        value['id'] = self.id
        return value


class ReportLog(OutputLog):
    """ Log of a SED report

    Attributes:
        id (:obj:`str`): id of the report
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        data_sets (:obj:`dict` of :obj:`str` to :obj:`Status`): execution status of each
            data set
        parent (:obj:`SedDocumentLog`): execution status of parent SED document
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None, data_sets=None,
                 parent=None, out_dir=None):
        """
        Args:
            id (:obj:`str`, optional): id of the report
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            data_sets (:obj:`dict` of :obj:`str` to :obj:`Status`, optional): execution status of each
                data set
            parent (:obj:`SedDocumentLog`, optional): execution status of parent SED document
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(ReportLog, self).__init__(id=id, status=status, exception=exception, skip_reason=skip_reason,
                                        output=output, duration=duration, parent=parent, out_dir=out_dir)
        self.data_sets = data_sets

    def finalize(self):
        """ Mark all unexceuted elements as skipped """
        super(ReportLog, self).finalize()

        if self.data_sets:
            for id, status in self.data_sets.items():
                if status == Status.QUEUED:
                    self.data_sets[id] = Status.SKIPPED
                elif status == Status.RUNNING:
                    self.data_sets[id] = Status.FAILED

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        dict_log = super(ReportLog, self).to_json()
        dict_log['dataSets'] = (
            [{'id': id, 'status': status.value if status else None} for id, status in self.data_sets.items()]
            if self.data_sets is not None
            else None
        )
        return dict_log


class Plot2DLog(OutputLog):
    """ Log of a 2D SED plot

    Attributes:
        id (:obj:`str`): id of the plot
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        curves (:obj:`dict` of :obj:`str` to :obj:`Status`): execution status of each
            curve
        parent (:obj:`SedDocumentLog`): execution status of parent SED document
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None, curves=None,
                 parent=None, out_dir=None):
        """
        Args:
            id (:obj:`str`, optional): id of the plot
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            curves (:obj:`dict` of :obj:`str` to :obj:`Status`, optional): execution status of each
                curve
            parent (:obj:`SedDocumentLog`, optional): execution status of parent SED document
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(Plot2DLog, self).__init__(id=id, status=status, exception=exception, skip_reason=skip_reason,
                                        output=output, duration=duration, parent=parent, out_dir=out_dir)
        self.curves = curves

    def finalize(self):
        """ Mark all unexceuted elements as skipped """
        super(Plot2DLog, self).finalize()

        if self.curves:
            for id, status in self.curves.items():
                if status == Status.QUEUED:
                    self.curves[id] = Status.SKIPPED
                elif status == Status.RUNNING:
                    self.curves[id] = Status.FAILED

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        dict_log = super(Plot2DLog, self).to_json()
        dict_log['curves'] = (
            [{'id': id, 'status': status.value if status else None} for id, status in self.curves.items()]
            if self.curves is not None
            else None
        )
        return dict_log


class Plot3DLog(OutputLog):
    """ Log of a 3D SED plot

    Attributes:
        id (:obj:`str`): id of the plot
        status (:obj:`Status`): execution status of the archive
        exception (:obj:`Exception`): exception
        skip_reason (:obj:`Exception`): reason of skip
        output (:obj:`str`): output
        duration (:obj:`float`): duration in seconds
        surfaces (:obj:`dict` of :obj:`str` to :obj:`Status`): execution status of each
            surface
        parent (:obj:`SedDocumentLog`): execution status of parent SED document
        out_dir (:obj:`str`): directory to export status
    """

    def __init__(self, id=None, status=None, exception=None, skip_reason=None, output=None, duration=None, surfaces=None,
                 parent=None, out_dir=None):
        """
        Args:
            id (:obj:`str`, optional): id of the report
            status (:obj:`Status`, optional): execution status of the archive
            exception (:obj:`Exception`, optional): exception
            skip_reason (:obj:`Exception`, optional): reason of skip
            output (:obj:`str`, optional): output
            duration (:obj:`float`, optional): duration in seconds
            surfaces (:obj:`dict` of :obj:`str` to :obj:`Status`, optional): execution status of each
                surface
            parent (:obj:`SedDocumentLog`, optional): execution status of parent SED document
            out_dir (:obj:`str`, optional): directory to export status
        """
        super(Plot3DLog, self).__init__(id=id, status=status, exception=exception, skip_reason=skip_reason,
                                        output=output, duration=duration, parent=parent, out_dir=out_dir)
        self.surfaces = surfaces

    def finalize(self):
        """ Mark all unexceuted elements as skipped """
        super(Plot3DLog, self).finalize()

        if self.surfaces:
            for id, status in self.surfaces.items():
                if status == Status.QUEUED:
                    self.surfaces[id] = Status.SKIPPED
                elif status == Status.RUNNING:
                    self.surfaces[id] = Status.FAILED

    def to_json(self):
        """ Generate a JSON-compatible representation

        Returns:
            :obj:`dict`: JSON-compatible representation
        """
        dict_log = super(Plot3DLog, self).to_json()
        dict_log['surfaces'] = (
            [{'id': id, 'status': status.value if status else None} for id, status in self.surfaces.items()]
            if self.surfaces is not None
            else None
        )
        return dict_log


class StandardOutputErrorCapturerLevel(int, enum.Enum):
    """ Level at which stdout/stderr should be captured """
    python = 0
    c = 1



""" Exceptions for COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-12
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..exceptions import BioSimulatorsException

__all__ = [
    'CombineArchiveExecutionError',
    'NoSedmlError',
]


class CombineArchiveExecutionError(BioSimulatorsException):
    """ Error that a SED document could not be executed """
    pass  # pragma: no cover


class NoSedmlError(BioSimulatorsException):
    """ Error that a COMBINE/OMEX archive does not contain any SED-ML files """
    pass  # pragma: no cover



""" Methods for validate BioSimulations metadata

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-06-23
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..combine.data_model import CombineArchiveContentFormatPattern
from ..utils.identifiers_org import validate_identifiers_org_uri, InvalidIdentifiersOrgUri
from .data_model import BIOSIMULATIONS_PREDICATE_TYPES, BIOSIMULATIONS_THUMBNAIL_FORMATS
from .utils import get_global_combine_archive_content_uri
import dateutil.parser
import os
import re
import validators

__all__ = [
    'validate_biosimulations_metadata',
    'validate_biosimulations_metadata_for_uri',
]


def validate_biosimulations_metadata(metadata, archive=None, working_dir=None):
    """ Validate BioSimulations metadata for a COMBINE/OMEX archive

    Args:
        metadata (:obj:`list` of :obj:`dict`): BioSimulations metadata about 1 or more URIs
        archive (:obj:`CombineArchive`, optional): parent COMBINE archive
        working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors with the metadata
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the metadata
    """
    errors = []
    warnings = []

    has_archive_metadata = False
    for el_metadata in metadata:
        el_is_archive = el_metadata['uri'] == '.' and el_metadata['combine_archive_uri']
        has_archive_metadata = has_archive_metadata or el_is_archive

        temp_errors, temp_warnings = validate_biosimulations_metadata_for_uri(
            el_metadata, validate_minimal_metadata=el_is_archive, archive=archive, working_dir=working_dir)

        if temp_errors:
            el_uri = get_global_combine_archive_content_uri(el_metadata['uri'], el_metadata['combine_archive_uri'])
            errors.append(['The metadata for URI `{}` is invalid.'.format(
                el_uri), temp_errors])

        if temp_warnings:
            el_uri = get_global_combine_archive_content_uri(el_metadata['uri'], el_metadata['combine_archive_uri'])
            warnings.append(['The metadata for URI `{}` may be invalid.'.format(
                el_uri),  temp_warnings])

    if not has_archive_metadata:
        errors.append([(
            'The metadata does not contain information about a parent COMBINE/OMEX archive '
            '(e.g., `rdf:about="http://omex-library.org/BioSim0001.omex"`). '
            'Archive-level metadata is required for publication to BioSimulations.'
        )])

    return errors, warnings


def validate_biosimulations_metadata_for_uri(metadata, validate_minimal_metadata=False, archive=None, working_dir=None):
    """ Validate BioSimulations metadata for a file in a COMBINE/OMEX archive

    Args:
        metadata (:obj:`dict`): BioSimulations metadata
        validate_minimal_metadata (:obj:`bool`, optional): whether to check that all required metadata attributes
            are defined
        archive (:obj:`CombineArchive`, optional): parent COMBINE archive
        working_dir (:obj:`str`, optional): working directory (e.g., directory of the parent COMBINE/OMEX archive)

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors with the metadata
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the metadata
    """
    errors = []
    warnings = []

    # required attributes are present
    if validate_minimal_metadata:
        for predicate_type in BIOSIMULATIONS_PREDICATE_TYPES.values():
            if predicate_type['required'] and (
                (not predicate_type['multiple_allowed'] and metadata[predicate_type['attribute']] is None)
                or (predicate_type['multiple_allowed'] and metadata[predicate_type['attribute']] == [])
            ):
                errors.append(['Attribute `{}` ({}) is required.'.format(
                    predicate_type['attribute'], predicate_type['uri'])])

    # URIs are URLs
    # Identifiers.org URLs point to valid identifiers
    for predicate_type in BIOSIMULATIONS_PREDICATE_TYPES.values():
        if predicate_type['has_uri'] and predicate_type['has_label']:
            if predicate_type['multiple_allowed']:
                objects = metadata[predicate_type['attribute']]
            else:
                objects = [metadata[predicate_type['attribute']]]

            for object in objects:
                if object and object['uri']:
                    if not validators.url(object['uri']):
                        errors.append(['URI `{}` of attribute `{}` ({}) is not a valid URI.'.format(
                            object['uri'], predicate_type['attribute'], predicate_type['uri'])])
                    else:
                        match = re.match(r'^https?://identifiers\.org/(([^/:]+/[^/:]+|[^/:]+)[/:](.+))$', object['uri'])
                        if match:
                            try:
                                validate_identifiers_org_uri(object['uri'])
                            except InvalidIdentifiersOrgUri as exception:
                                msg = (
                                    'URI `{}` of attribute `{}` ({}) is not a valid Identifiers.org identifier.'
                                ).format(object['uri'], predicate_type['attribute'], predicate_type['uri'])
                                errors.append([msg, [[str(exception)]]])

    # thumbnail is a file; file type is checked by COMBINE validation
    if working_dir:
        for thumbnail in metadata['thumbnails']:
            thumbnail = os.path.relpath(thumbnail, '.')
            thumbnail_filename = os.path.join(working_dir, thumbnail)
            if os.path.isfile(thumbnail_filename):
                if archive:
                    for content in archive.contents:
                        if (
                            content
                            and content.location
                            and thumbnail == os.path.relpath(content.location, '.')
                        ):
                            is_valid = False
                            for format in BIOSIMULATIONS_THUMBNAIL_FORMATS:
                                if content.format and re.match(CombineArchiveContentFormatPattern[format], content.format):
                                    is_valid = True
                                    break
                            if not is_valid:
                                errors.append(['The format of thumbnail `{}` must be one of the following:'.format(thumbnail),
                                               sorted([[format] for format in BIOSIMULATIONS_THUMBNAIL_FORMATS])])

            else:
                errors.append(['Thumbnail `{}` is not a file.'.format(thumbnail)])
    else:
        if metadata['thumbnails']:
            warnings.append([('The locations of the thumbnails could not be validated '
                              'because a working directory was not provided.')])

    # created is a date
    if metadata['created'] is not None:
        try:
            dateutil.parser.parse(metadata['created'])
        except dateutil.parser.ParserError:
            errors.append(['Created date `{}` is not a valid date.'.format(metadata['created'])])

    # modified are dates
    for date in metadata['modified']:
        try:
            dateutil.parser.parse(date)
        except dateutil.parser.ParserError:
            errors.append(['Modified date `{}` is not a valid date.'.format(date)])

    # return errors and warnings
    return (errors, warnings)



""" Data model for GitHub action workflows

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-08
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..exceptions import BioSimulatorsException

__all__ = [
    'GitHubActionCaughtError',
    'Comment',
]


class GitHubActionCaughtError(BioSimulatorsException):
    """ An error caught during the execution of a GitHub action """
    pass  # pragma: no cover


class Comment(object):
    """ A comment on an issue

    Attributes:
        text (:obj:`str`): text
        error (:obj:`bool`): if :obj:`True`, format as error
    """

    def __init__(self, text=None, error=None):
        """
        Args:
            text (:obj:`str`): text
            error (:obj:`bool`): if :obj:`True`, format as error
        """
        self.text = text
        self.error = error



""" Utility methods for working with BioSimulations and runBioSimulations

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-30
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""
from ..config import get_config, Config  # noqa: F401
import json
import os
import requests
import simplejson.errors

__all__ = [
    'submit_project_to_runbiosimulations',
    'validate_biosimulations_api_response',
]


def submit_project_to_runbiosimulations(name, filename_or_url,
                                        simulator, simulator_version='latest',
                                        cpus=1, memory=8, max_time=20, env_vars=None,
                                        email=None, public=False,
                                        config=None):
    """ Submit a simulation project (COMBINE/OMEX archive) to runBioSimulations and, optionally, BioSimulations

    Args:
        name (:obj:`str`): name for the simulation run
        filename_or_url (:obj:`str`): path or URL to COMBINE/OMEX archive
        simulator (:obj:`str`): BioSimulators id for simulator
        simulator_version (:obj:`str`, optional): simulator version
        cpus (:obj:`int`, optional): CPUs
        memory (:obj:`float`, optional): maximum memory in GB
        max_time (:obj:`float`, optional): maximum execution time in minutes
        env_vars (:obj:`list` of :obj:`dict`, optional): environment variables to execute the COMBINE/OMEX archive. Each
            element should have two string-valued keys ``key`` and ``value``
        email (:obj:`str`, optional): email to receive a notification upon completion of the simulation run
        public (:obj:`bool`, optional): whether to publish the simulation run to BioSimulations
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`str`: runBioSimulations id
    """
    config = config or get_config()
    endpoint = config.RUNBIOSIMULATIONS_API_ENDPOINT + 'runs'
    run = {
        "name": name,
        "simulator": simulator,
        "simulatorVersion": simulator_version,
        "cpus": cpus,
        "memory": memory,
        "maxTime": max_time,
        "envVars": env_vars or [],
        "email": email,
        "public": public,
    }
    if os.path.isfile(filename_or_url):
        with open(filename_or_url, 'rb') as file:
            response = requests.post(
                endpoint,
                data={
                    'simulationRun': json.dumps(run),
                },
                files={
                    'file': ('project.omex', file, 'application/zip'),
                },
                headers={
                    "Accept": "application/json",
                }
            )
    else:
        run['url'] = filename_or_url
        response = requests.post(
            endpoint,
            json=run,
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        )
    response.raise_for_status()
    return response.json()['id']


def validate_biosimulations_api_response(response, failure_introductory_message):
    """ Validate a response from one of BioSimulation's APIs.

    Args:
        response (:obj:`requests.models.Response`): API response
        failure_introductory_message (:obj:`str`): introductory message for failures

    Raises:
        :obj:`requests.RequestException`
    """
    try:
        response.raise_for_status()

    except requests.RequestException as exception:
        try:
            errors = response.json()['error']
        except (simplejson.errors.JSONDecodeError, KeyError):
            raise exception

        error_messages = []
        for error in errors:
            pointer = error.get('source', {}).get('pointer', None)
            error_messages.append('{} ({}):{}{}'.format(
                error['title'], error['status'],
                ' `{}` is invalid'.format(pointer) if pointer else '',
                '\n\n  ' + error['detail'].replace('\n', '\n  ') if error.get('detail', None) else '',
            ))

        error_message = '{}\n\n  - {}'.format(
            failure_introductory_message,
            '\n\n  - '.join(msg.replace('\n', '\n    ') for msg in error_messages))

        exception.args = (error_message,)

        raise
